import argparse

from franken.backbones import REGISTRY
from franken.backbones.utils import download_checkpoint, make_summary
from franken.utils.misc import setup_logger


### Command 'list': list available models


def build_list_arg_parser(subparsers) -> argparse.ArgumentParser:
    parser = subparsers.add_parser("list", description="List available models")
    parser.add_argument(
        "--cache_dir",
        help=(
            "Directory to save the downloaded checkpoints. "
            "Defaults to '~/.franken/' in the user home or to the "
            "'FRANKEN_CACHE_DIR' environment variable if set."
        ),
        type=str,
        default=None,
    )
    parser.add_argument(
        "--log_level",
        help="log-level for the command-line logger",
        type=str,
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
    )
    return parser


def run_list_cmd(args):
    setup_logger(level=args.log_level, directory=None)
    print(make_summary(cache_dir=args.cache_dir))


### Command 'download': download a model


def build_download_arg_parser(subparsers) -> argparse.ArgumentParser:
    parser = subparsers.add_parser("download", description="Download a model")
    parser.add_argument(
        "--model_name",
        help="The name of the model to download.",
        type=str,
        required=True,
        choices=[
            name for name, info in REGISTRY.items() if info["implemented"] is True
        ],
    )
    parser.add_argument(
        "--cache_dir",
        help=(
            "Directory to save the downloaded checkpoints. "
            "Defaults to '~/.franken/' in the user home or to the "
            "'FRANKEN_CACHE_DIR' environment variable if set."
        ),
        type=str,
        default=None,
    )
    parser.add_argument(
        "--log_level",
        help="log-level for the command-line logger",
        type=str,
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
    )
    return parser


def run_download_cmd(args):
    setup_logger(level=args.log_level, directory=None)
    download_checkpoint(args.model_name, args.cache_dir)


def build_arg_parser():
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(
        required=True,
        title="Franken backbone CLI",
        description="Provides helpers to interact with the various backbone models supported by Franken",
        help="Run `%(prog)s <subcommand> -h` for help with the individual subcommands",
    )

    list_parser = build_list_arg_parser(subparsers)
    list_parser.set_defaults(func=run_list_cmd)
    download_parser = build_download_arg_parser(subparsers)
    download_parser.set_defaults(func=run_download_cmd)

    return parser


def main():
    """This entry-point has 2 commands, 'list' and 'download'.
    Usage:
        franken.backbones list <args>
        franken.backbones download <args>
    """
    parser = build_arg_parser()
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



import json
import logging
import pathlib
from unittest.mock import DEFAULT, MagicMock, Mock, mock_open, patch

import pytest
import torch

from franken.data.base import Configuration, Target
from franken.rf.model import FrankenPotential
from franken.rf.scaler import Statistics
from franken.trainers import RandomFeaturesTrainer
from franken.trainers.log_utils import DataSplit, LogCollection, LogEntry
from tests.utils import mocked_gnn
from .conftest import DEVICES


@pytest.mark.parametrize("device", DEVICES)
class TestTrainer:
    def trainer(self, device, dtype=torch.float32, **kwargs):
        return RandomFeaturesTrainer(
            # it's okay to pass None here as long as we don't test `fit` or `set_statistics`
            train_dataloader=None, # type: ignore
            random_features_normalization=None,
            log_dir=None,
            save_every_model=False,
            device=device,
            dtype=dtype,
            **kwargs
        )

    def gen_efmap_vals(self, num_feat, num_configs, num_atoms, device, dtype):
        return [
            (
                torch.randn(num_feat, num_atoms * 3, device=device, dtype=dtype),
                torch.randn(num_feat, device=device, dtype=dtype),
            )
            for _ in range(num_configs)
        ]

    def gen_data(self, num_configs, num_atoms, dtype, split="train"):
        data = [
            (
                Configuration(torch.randn(num_atoms, 3, dtype=dtype), torch.randint(1, 100, (num_atoms, )), torch.tensor(num_atoms)),
                Target(torch.randn(1) ** 2, torch.randn(num_atoms, 3) ** 2)
            )
            for _ in range(num_configs)
        ]
        def iterator(*args, **kwargs):
            for x in data:
                yield x
        dl_mock = MagicMock()
        dl_mock.__len__ = MagicMock(return_value=num_configs)
        dl_mock.__iter__ = iterator
        dl_mock.dataset.__len__ = MagicMock(return_value=num_configs)
        dl_mock.dataset.split = split
        return dl_mock

    def init_model(self, efmaps, device, dtype):
        # We need the following from the potential model
        # - model.rf.total_random_features
        # - model.grad_feature_map(data)
        with mocked_gnn(device=device, dtype=dtype, feature_dim=32):
            underlying = FrankenPotential(
                    gnn_backbone_id="test",
                    kernel_type="gaussian",
                    random_features_params={"num_random_features": 64},
                ).to(device)
            model = Mock(wraps=underlying)
            model.rf = Mock(wraps=underlying.rf)
            if efmaps is not None:
                # will iterate through efmap_vals at every call
                model.grad_feature_map = MagicMock(side_effect=efmaps)
                model.rf.total_random_features = efmaps[0][0].shape[0]
            else:
                model.rf.total_random_features = underlying.rf.total_random_features
            return model

    def test_covs_and_coeffs(self, device):
        trainer = self.trainer(device, torch.float32)
        num_configs = 3
        num_atoms = 10
        num_rf = 4
        efmap_vals = self.gen_efmap_vals(num_rf, num_configs, num_atoms, device, trainer.buffer_dt)
        dataloader = self.gen_data(num_configs, num_atoms, trainer.buffer_dt)
        # Expected energy and forces covariance and coeffs
        exp_f_cov = sum([efmap[0] @ efmap[0].T for efmap in efmap_vals])
        exp_f_coef = sum([-efmap[0] @ data[1].forces.to(device).view(-1) / num_atoms for efmap, data in zip(efmap_vals, dataloader)])
        exp_e_cov = sum([torch.outer(efmap[1], efmap[1]) for efmap in efmap_vals])
        exp_e_coef = sum([efmap[1] * data[1].energy.to(device) / num_atoms for efmap, data in zip(efmap_vals, dataloader)])

        model = self.init_model(efmap_vals, device, trainer.buffer_dt)
        trainer._compute_covs_and_coeffs(model, dataloader)
        # Check outputs
        torch.testing.assert_close(
            torch.triu(trainer.covariance, 1), torch.triu(exp_e_cov, 1) # type: ignore
        )
        torch.testing.assert_close(
            trainer.diag_energy, torch.diagonal(exp_e_cov) # type: ignore
        )
        torch.testing.assert_close(
            torch.tril(trainer.covariance, -1), torch.tril(exp_f_cov, -1) # type: ignore
        )
        torch.testing.assert_close(
            trainer.diag_forces, torch.diagonal(exp_f_cov) # type: ignore
        )
        torch.testing.assert_close(
            trainer.coeffs_energy, exp_e_coef
        )
        torch.testing.assert_close(
            trainer.coeffs_forces, exp_f_coef
        )
        # Check behavior
        assert model.grad_feature_map.call_count == num_configs

    def test_solve(self, device):
        # Start by computing covs and coeffs
        trainer = self.trainer(device, torch.float32)
        num_configs = 3
        num_atoms = 10
        num_rf = 32
        loss_lerp_weight = 0.1
        L2_penalty = 1e-6
        efmap_vals = self.gen_efmap_vals(num_rf, num_configs, num_atoms, device, trainer.buffer_dt)
        dataloader = self.gen_data(num_configs, num_atoms, trainer.buffer_dt)
        # Expected energy and forces covariance and coeffs
        exp_f_cov = sum([efmap[0] @ efmap[0].T for efmap in efmap_vals])
        exp_f_coef = sum([-efmap[0] @ data[1].forces.to(device).view(-1) / num_atoms for efmap, data in zip(efmap_vals, dataloader)])
        exp_e_cov = sum([torch.outer(efmap[1], efmap[1]) for efmap in efmap_vals])
        exp_e_coef = sum([efmap[1] * data[1].energy.to(device) / num_atoms for efmap, data in zip(efmap_vals, dataloader)])
        lerped_cov = exp_f_cov * loss_lerp_weight + exp_e_cov * (1 - loss_lerp_weight)
        exp_sol = torch.linalg.solve(
            lerped_cov + torch.eye(num_rf, dtype=trainer.buffer_dt, device=device) * L2_penalty,
            exp_f_coef * loss_lerp_weight + exp_e_coef * (1 - loss_lerp_weight)
        )
        model = self.init_model(efmap_vals, device, trainer.buffer_dt)
        trainer._compute_covs_and_coeffs(model, dataloader)
        solution = trainer._solve(loss_lerp_weight=loss_lerp_weight, L2_penalty=L2_penalty)
        torch.testing.assert_close(exp_sol, solution)

    def test_compute_train_preds_shortcut(self, device):
        trainer = self.trainer(device, torch.float64, save_fmaps=True)
        num_configs = 5
        num_atoms = 10
        dataloader = self.gen_data(num_configs, num_atoms, trainer.buffer_dt, split="train")
        model = self.init_model(None, device, trainer.buffer_dt)
        trainer._compute_covs_and_coeffs(model, dataloader)
        weights = trainer._solve(loss_lerp_weight=0.3, L2_penalty=1e-3)
        weights = weights.view(1, -1)  # fit converts to this shape

        logs = LogCollection([LogEntry("", 0, 0, 0, [], [])])
        logs_fast = trainer.evaluate(model, dataloader, logs, weights, ["forces_MAE", "energy_MAE"])
        model.energy_and_forces.assert_not_called()
        model.energy_and_forces_from_fmaps.assert_called()
        model.energy_and_forces_from_fmaps.reset_mock()

        trainer.save_fmaps = False
        logs = LogCollection([LogEntry("", 0, 0, 0, [], [])])
        logs_slow = trainer.evaluate(model, dataloader, logs, weights, ["forces_MAE", "energy_MAE"])
        model.energy_and_forces.assert_called()
        model.energy_and_forces_from_fmaps.assert_not_called()

        metrics_fast = logs_fast[0].metrics
        metrics_slow = logs_slow[0].metrics
        for mfast, mslow in zip(metrics_fast, metrics_slow):
            assert mfast.name == mslow.name
            torch.testing.assert_close(mslow.value, mfast.value, msg=f"Slow and fast computations don't match metric {mfast.name}")


class TestSerializeBestModel:
    @pytest.fixture
    def log_collection(self):
        return LogCollection(logs=[LogEntry.from_dict(log) for log in [{
            "checkpoint": {"hash": "rand_uuid", "rf_weight_id": 0},
            "timings": {"cov_coeffs": 1.0, "solve": 1.0},
            "metrics": {
                "train": {"energy_MAE": 1.0, "forces_MAE": 1.0, "forces_cosim": 1.0},
                "validation": {"energy_MAE": 1.0, "forces_MAE": 1.0, "forces_cosim": 1.0},
                "test": {"energy_MAE": 1.0, "forces_MAE": 1.0, "forces_cosim": 1.0},
            },
            "hyperparameters": {
                "franken": {
                    "gnn_backbone_id": "SchNet-S2EF-OC20-All",
                    "interaction_block": 3,
                    "kernel_type": "gaussian",
                },
                "random_features": {
                    "num_random_features": 1024,
                },
                "input_scaler": {"scale_by_Z": True, "num_species": 2},
                "solver": {
                    "L2_penalty": 1e-6,
                    "loss_lerp_weight": 0.1,
                    "dtype": "torch.float64",
                },
            },
        },{
            "checkpoint": {"hash": "rand_uuid2", "rf_weight_id": 1},
            "timings": {"cov_coeffs": 1.0, "solve": 1.0},
            "metrics": {
                "train": {"energy_MAE": 1.0, "forces_MAE": 1.0, "forces_cosim": 1.0},
                "validation": {"energy_MAE": 1.0, "forces_MAE": 0.5, "forces_cosim": 1.0},
                "test": {"energy_MAE": 1.0, "forces_MAE": 1.0, "forces_cosim": 1.0},
            },
            "hyperparameters": {
                "franken": {
                    "gnn_backbone_id": "SchNet-S2EF-OC20-All",
                    "interaction_block": 3,
                    "kernel_type": "gaussian",
                },
                "random_features": {
                    "num_random_features": 1024,
                },
                "input_scaler": {"scale_by_Z": True, "num_species": 2},
                "solver": {
                    "L2_penalty": 1e-6,
                    "loss_lerp_weight": 0.1,
                    "dtype": "torch.float64",
                },
            },
        }]])

    def test_should_save(self, log_collection, caplog):
        with patch.object(pathlib.Path, 'exists') as mock_exists:
            with mocked_gnn("cpu", torch.float32):
                with patch.multiple("franken.trainers.base", LogCollection=DEFAULT) as p2:
                    with patch.multiple("torch", load=DEFAULT, save=DEFAULT) as p3:
                        with patch("builtins.open", mock_open()) as file_open:
                            trainer = RandomFeaturesTrainer(
                                train_dataloader=None, # type: ignore
                                random_features_normalization=None,
                                log_dir=pathlib.Path("test_dir"),
                                save_every_model=False,
                                device="cpu",
                                dtype=torch.float32
                            )
                            model = FrankenPotential(
                                gnn_backbone_id="test",
                                kernel_type="gaussian",
                            ).to("cpu")
                            all_weights = torch.randn(2, 1024, dtype=trainer.buffer_dt)
                            # 1. best_model_file exists but contains same best model as log_collection
                            mock_exists.return_value = True
                            p2["LogCollection"].from_json.return_value = log_collection
                            # log_collection[1] is best model (read will be used to open best_model_file)
                            file_open.return_value.read.return_value = json.dumps(log_collection[1].to_dict())
                            with caplog.at_level(logging.DEBUG):
                                trainer.serialize_best_model(
                                    model=model,
                                    all_weights=all_weights,
                                    split=DataSplit.VALIDATION
                                )
                            p3["save"].assert_not_called()
                            assert caplog.text == ""
                            # 2. best_model_file exists and does not contain same best model as log_collection
                            mock_exists.return_value = True
                            p2["LogCollection"].from_json.return_value = log_collection
                            # log_collection[0] is not best model (read will be used to open best_model_file)
                            file_open.return_value.read.return_value = json.dumps(log_collection[0].to_dict())
                            with caplog.at_level(logging.DEBUG):
                                trainer.serialize_best_model(
                                    model=model,
                                    all_weights=all_weights,
                                    split=DataSplit.VALIDATION
                                )
                            p3["save"].assert_called_once()
                            assert "Identified new best model" in caplog.text
                            assert "Saved best model" in caplog.text
                            assert p3["save"].call_args.args[1] == pathlib.Path("test_dir/best_ckpt.pt")
                            p3["save"].reset_mock()
                            # 3. best_model_file does not exist
                            mock_exists.return_value = False
                            p2["LogCollection"].from_json.return_value = log_collection
                            with caplog.at_level(logging.DEBUG):
                                trainer.serialize_best_model(
                                    model=model,
                                    all_weights=all_weights,
                                    split=DataSplit.VALIDATION
                                )
                            p3["save"].assert_called_once()
                            assert "Identified new best model" in caplog.text
                            assert "Saved best model" in caplog.text
                            assert p3["save"].call_args.args[1] == pathlib.Path("test_dir/best_ckpt.pt")

    @pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
    def test_dtype_correct(self, dtype, log_collection):
        with patch.object(pathlib.Path, 'exists') as mock_exists:
            mock_exists.return_value = False
            trainer = RandomFeaturesTrainer(
                train_dataloader=None, # type: ignore
                random_features_normalization=None,
                log_dir=pathlib.Path("test_dir"),
                save_every_model=False,
                device="cpu",
                dtype=dtype
            )
            all_weights = torch.randn(2, 1024, dtype=trainer.buffer_dt)
            with mocked_gnn("cpu", torch.float32):
                with patch.multiple("franken.trainers.base", LogCollection=DEFAULT) as p2:
                    with patch.multiple("torch", load=DEFAULT, save=DEFAULT) as p3:
                        with patch("builtins.open", mock_open()):
                            p2["LogCollection"].from_json.return_value = log_collection
                            model = FrankenPotential(
                                gnn_backbone_id="test",
                                kernel_type="gaussian",
                            ).to("cpu")
                            trainer.serialize_best_model(
                                model=model,
                                all_weights=all_weights,
                                split=DataSplit.VALIDATION
                            )
                            # check model saved with correct parameters.
                            assert p3["save"].call_args.args[1] == pathlib.Path("test_dir/best_ckpt.pt")
                            save_ckpt = p3["save"].call_args.args[0]
                            assert save_ckpt["multi_weights"] is None
                            # check model weights. This checks dtype!
                            torch.testing.assert_close(model.rf.weights, all_weights[1].view(1, -1))


def test_get_statistics():
    dloader = Mock()
    dloader.dataset = Mock()
    with patch.multiple("franken.trainers.base", compute_dataset_statistics=DEFAULT) as p2:
        p2["compute_dataset_statistics"].return_value = Statistics()
        trainer = RandomFeaturesTrainer(
            train_dataloader=dloader, # type: ignore
            random_features_normalization=None,
            log_dir=pathlib.Path("test_dir"),
            save_every_model=False,
            device="cpu",
            dtype=torch.float32
        )
        with mocked_gnn("cpu", torch.float32):
            model = FrankenPotential(
                gnn_backbone_id="test",
                kernel_type="gaussian",
                random_gnn_arg=32,  # will be stored in gnn.init_args()
            ).to("cpu")
            trainer.get_statistics(model)
            p2["compute_dataset_statistics"].assert_called_once()
            p2["compute_dataset_statistics"].reset_mock()
            trainer.get_statistics(model)
            p2["compute_dataset_statistics"].assert_not_called()
            # Initialize a new GNN (different parameters)
            model = FrankenPotential(
                gnn_backbone_id="test2",
                kernel_type="gaussian",
                random_gnn_arg=32,  # will be stored in gnn.init_args()
            ).to("cpu")
            trainer.get_statistics(model)
            p2["compute_dataset_statistics"].assert_called_once()



from pathlib import Path
from typing import Literal, Union

import numpy as np
import torch

from ase.calculators.calculator import Calculator, all_changes
from franken.data import BaseAtomsDataset, Configuration
from franken.rf.model import FrankenPotential
from franken.utils.misc import get_device_name


class FrankenCalculator(Calculator):
    implemented_properties = ["energy", "forces"]
    default_parameters = {}
    nolabel = True  # ??

    def __init__(
        self,
        franken_ckpt: Union[FrankenPotential, str, Path],
        device=None,
        rf_weight_id: int | None = None,
        forces_mode: Literal["torch.func", "torch.autograd"] = "torch.autograd",
        **calc_kwargs,
    ):
        super().__init__(**calc_kwargs)
        self.franken: FrankenPotential
        if isinstance(franken_ckpt, torch.nn.Module):
            self.franken = franken_ckpt
            if device is not None:
                self.franken.to(device)
        else:
            # Handle jitted torchscript archives and normal files
            try:
                self.franken = torch.jit.load(franken_ckpt, map_location=device)
            except RuntimeError as e:
                if "PytorchStreamReader" not in str(e):
                    raise
                self.franken = FrankenPotential.load(  # type: ignore
                    franken_ckpt,
                    map_location=device,
                    rf_weight_id=rf_weight_id,
                )

        self.dataset = BaseAtomsDataset.from_path(
            data_path=None,
            split="md",
            gnn_backbone_id=self.franken.gnn_backbone_id,
        )
        self.device = (
            device if device is not None else next(self.franken.parameters()).device
        )
        self.forces_mode = forces_mode

    def calculate(
        self,
        atoms=None,
        properties=None,
        system_changes=all_changes,
    ):
        if properties is None:
            properties = self.implemented_properties
        if "forces" not in properties:
            forces_mode = "no_forces"
        else:
            forces_mode = self.forces_mode

        super().calculate(atoms, properties, system_changes)

        # self.atoms is set in the super() call. Unclear why it should be preferred over `atoms`
        config_idx = self.dataset.add_configuration(self.atoms)  # type: ignore
        cpu_data = self.dataset.__getitem__(config_idx, no_targets=True)
        assert isinstance(cpu_data, Configuration)
        data = cpu_data.to(self.device)

        energy, forces = self.franken.energy_and_forces(data, forces_mode=forces_mode)

        if energy.ndim == 0:
            self.results["energy"] = energy.item()
        else:
            self.results["energy"] = np.squeeze(energy.numpy(force=True))
        if "forces" in properties:
            assert forces is not None
            self.results["forces"] = np.squeeze(forces.numpy(force=True))


def calculator_throughput(
    calculator, atoms_list, num_repetitions=1, warmup_configs=5, verbose=True
):
    from time import perf_counter

    hardware = get_device_name(calculator.device)

    _atom_numbers = set(len(atoms) for atoms in atoms_list)
    assert (
        len(_atom_numbers) == 1
    ), f"This function only accepts configurations with the same number of atoms, while found configurations with {_atom_numbers} number of atoms"
    natoms = _atom_numbers.pop()

    assert len(atoms_list) > warmup_configs
    for idx in range(warmup_configs):
        calculator.calculate(atoms_list[idx])
    time_init = perf_counter()
    for _ in range(num_repetitions):
        for atoms in atoms_list:
            calculator.calculate(atoms)
    time = perf_counter() - time_init
    configs_per_sec = (len(atoms_list) * num_repetitions) / time
    results = {
        "throughput": configs_per_sec,
        "atoms": natoms,
        "hardware": hardware,
    }
    if verbose:
        print(
            f"{results['throughput']:.1f} cfgs/sec ({results['atoms']} atoms) | {results['hardware']}"
        )
    return results



from pathlib import Path
from typing import Optional

import torch
from mace.data import AtomicData
from mace.data.utils import config_from_atoms
from mace.tools.utils import AtomicNumberTable
from tqdm import tqdm

import franken.utils.distributed as dist_utils
from franken.backbones import REGISTRY
from franken.backbones.utils import get_checkpoint_path
from franken.data import BaseAtomsDataset, Configuration, Target


class MACEAtomsToGraphs:
    def __init__(self, z_table: AtomicNumberTable, cutoff: float):
        self.cutoff = cutoff
        self.z_table = z_table

    def convert(self, atoms):
        mace_config = config_from_atoms(atoms)
        graph = AtomicData.from_config(
            mace_config, z_table=self.z_table, cutoff=self.cutoff
        )
        graph.atomic_numbers = torch.tensor(atoms.get_atomic_numbers()).int()
        return graph

    def convert_all(
        self,
        atoms_list,
        process_rank: Optional[int] = None,
        split_name: Optional[str] = None,
    ):
        graphs = []
        atoms_iter = atoms_list
        if process_rank is None:
            process_rank = dist_utils.get_rank()
        if process_rank == 0:
            desc = "ASE -> MACE"
            if split_name is not None:
                desc += f" ({split_name})"
            atoms_iter = tqdm(atoms_list, desc=desc)
        for atoms in atoms_iter:
            graphs.append(self.convert(atoms))
        return graphs


class MACEAtomsDataset(BaseAtomsDataset):
    def __init__(
        self,
        data_path: str | Path | None,
        split: str,
        num_random_subsamples: int | None = None,
        subsample_rng: int | None = None,
        gnn_backbone_id: str | None = None,
        z_table: AtomicNumberTable | None = None,
        cutoff=6.0,
        precompute=True,
    ):
        super().__init__(data_path, split, num_random_subsamples, subsample_rng)
        if gnn_backbone_id is not None:
            z_table, cutoff = self.load_info_from_gnn_config(gnn_backbone_id)
        else:
            assert z_table is not None

        self.a2g = MACEAtomsToGraphs(z_table, cutoff)
        self.graphs = None
        if precompute and len(self.ase_atoms) > 0:
            self.graphs = self.a2g.convert_all(
                self.ase_atoms,
                split_name=self.split,
            )

    def load_info_from_gnn_config(self, gnn_backbone_id: str):
        ckpt_path = get_checkpoint_path(gnn_backbone_id)
        backbone_family = REGISTRY[gnn_backbone_id]["kind"].lower()
        if backbone_family != "mace":
            raise ValueError(f"Incompatible backbone family {backbone_family}")
        mace_gnn = torch.load(ckpt_path, map_location="cpu", weights_only=False)
        z_table = AtomicNumberTable([z.item() for z in mace_gnn.atomic_numbers])
        cutoff = mace_gnn.r_max.item()
        del mace_gnn
        return z_table, cutoff

    def __getitem__(self, idx, no_targets: bool = False):
        """Returns an array of (inputs, outputs) with inputs being a configuration
        and outputs being the target (energy and forces).
        Note: ONLY for the 'train' split, the energy_shift is removed from the target.
        """
        if self.graphs is None:
            graph = self.a2g.convert(self.ase_atoms[idx])
        else:
            graph = self.graphs[idx]

        data = Configuration(
            atom_pos=graph.positions,
            atomic_numbers=graph.atomic_numbers,
            natoms=torch.tensor(len(graph.atomic_numbers)).view(1),
            node_attrs=graph.node_attrs,
            edge_index=graph.edge_index,
            shifts=graph.shifts,
            unit_shifts=graph.unit_shifts,
        )
        if no_targets:
            return data

        energy = torch.tensor(
            self.ase_atoms[idx].get_potential_energy(apply_constraint=False)
        )
        if self.split == "train":
            energy = energy - self.energy_shifts[idx]

        target = Target(
            energy=energy,
            forces=torch.Tensor(self.ase_atoms[idx].get_forces(apply_constraint=False)),
        )
        return data, target



import sys
import numpy as np
from ase.io import read

def mad(data, axis=None):
  """
  Calculates the Mean Absolute Deviation (MAD) of a dataset.

  Args:
    data: The input data as a NumPy array or list.
    axis: The axis along which to compute the MAD. 
          If None, compute the MAD over the entire array.

  Returns:
    The Mean Absolute Deviation of the data.
  """
  return np.mean(np.absolute(data - np.mean(data, axis)), axis)

def analyze_trajectory(filename):
  """
  Calculates and prints the MAD of energy and forces for an ASE trajectory.
  Saves the results to a file.

  Args:
    filename: Path to the .xyz trajectory file.
    output_filename: Name of the output file to save the results.
  """
  traj = read(filename, index=":")  # Read the entire trajectory

  energies = np.array([atoms.get_potential_energy() for atoms in traj])
  forces = np.concatenate([atoms.get_forces() for atoms in traj])

  mad_energy = mad(energies)
  mad_forces = mad(forces)
  print("Warning: Script assumes energy values are in eV and Forces in eV/Å, which is the main convention in franken, otherwise ignore units.")
  print(f"Mean Absolute Deviation of Energy: {mad_energy:.4f} eV")
  print(f"Mean Absolute Deviation of Forces: {mad_forces:.4f} eV/Å")
  output_filename = filename.split(".")[0]
  # Save results to file
  with open(output_filename+"_metrics.txt", "w") as f:
    f.write("# Property\tMAD\n")
    f.write(f"Energy (eV)\t{mad_energy:.4f}\n")
    f.write(f"Forces (eV/Å)\t{mad_forces:.4f}\n")

if __name__ == "__main__":
  if len(sys.argv) < 2:
    print("Usage: python script.py <trajectory_file.xyz>")
    sys.exit(1)

  filename = sys.argv[1]
  analyze_trajectory(filename)


import os
from pathlib import Path
from unittest.mock import patch

import pytest

import franken.backbones.utils


@pytest.fixture
def mock_registry():
    return {
        "UNIMPLEMENTED_MODEL": {
            "kind": "mock",
            "implemented": False,
            "local": "unimplemented.ckpt",
            "remote": "https://example.com",
        }
    }


@pytest.fixture
def mock_cache_folder():
    return Path("/tmp/cache")


def test_model_registry():
    registry = franken.backbones.utils.load_model_registry()
    for model in registry.values():
        for key in ["remote", "local", "kind", "implemented"]:
            assert key in model.keys()


def test_cache_dir_default(mock_cache_folder):
    """Test that the function returns the default path when FRANKEN_CACHE_DIR is not set."""
    # Ensure no environment variable is set
    with patch.dict(os.environ, {}, clear=True):
        # Mock the home path and the Path.exists method
        with patch("pathlib.Path.home", return_value=mock_cache_folder):
            with patch("pathlib.Path.exists", return_value=True) as mock_exists:
                # Call the function
                franken.backbones.utils.CacheDir.initialize()
                result = franken.backbones.utils.CacheDir.get()

                # Check the default path is returned
                assert result == mock_cache_folder / ".franken"
                # Ensure that the path exists
                mock_exists.assert_called_once()


def test_cache_dir_with_env_var(mock_cache_folder):
    """Test that the function returns the correct path when FRANKEN_CACHE_DIR is set."""
    # Mock the environment variable
    with patch.dict(os.environ, {"FRANKEN_CACHE_DIR": str(mock_cache_folder)}):
        # Mock the Path.exists method
        with patch("pathlib.Path.exists", return_value=True) as mock_exists:
            # Call the function
            franken.backbones.utils.CacheDir.initialize()
            result = franken.backbones.utils.CacheDir.get()

            # Check the environment variable path is returned
            assert str(result) == str(mock_cache_folder)
            # Ensure that the path exists
            mock_exists.assert_called_once()


def test_download_checkpoint_name_error():
    """Test that a NameError is raised for unknown gnn_backbone_id."""
    # Mock the model registry to return an empty registry
    with patch("franken.backbones.utils.load_model_registry", return_value={}):
        # Expect a NameError when the gnn_backbone_id is not in the registry
        with pytest.raises(NameError) as exc_info:
            franken.backbones.utils.download_checkpoint("UNKNOWN_MODEL")
        assert "Unknown UNKNOWN_MODEL GNN backbone" in str(exc_info.value)


def test_download_checkpoint_not_implemented(mock_registry):
    """Test that a NotImplementedError is raised when the model is not implemented."""
    # Mock the model registry to return a registry with a model that is not implemented
    with patch(
        "franken.backbones.utils.load_model_registry", return_value=mock_registry
    ):
        # Expect a NotImplementedError when the gnn_backbone_id is not implemented
        with pytest.raises(NotImplementedError) as exc_info:
            franken.backbones.utils.download_checkpoint("UNIMPLEMENTED_MODEL")
        assert "The model UNIMPLEMENTED_MODEL is not implemented" in str(exc_info.value)


@pytest.mark.skip(reason="Actually downloads the model")
def test_download_checkpoint_successful_download(tmp_path):
    gnn_id = "MACE-L0"
    """Test that the model is downloaded correctly when it is implemented."""
    registry = franken.backbones.utils.load_model_registry()
    with patch.dict(os.environ, {"FRANKEN_CACHE_DIR": str(tmp_path)}):
        franken.backbones.utils.download_checkpoint(gnn_id)
        ckpt = tmp_path / "gnn_checkpoints" / registry[gnn_id]["local"]
        assert ckpt.exists()
        assert ckpt.is_file()


def test_get_checkpoint_path_valid_backbone(mock_registry, mock_cache_folder):
    with patch(
        "franken.backbones.utils.load_model_registry", return_value=mock_registry
    ), patch(
        "franken.backbones.utils.CacheDir.get", return_value=mock_cache_folder
    ), patch("pathlib.Path.exists", return_value=True):
        result = franken.backbones.utils.get_checkpoint_path("UNIMPLEMENTED_MODEL")
        expected_path = mock_cache_folder / "gnn_checkpoints" / "unimplemented.ckpt"
        assert result == expected_path


def test_get_checkpoint_path_invalid_backbone(mock_registry):
    with patch(
        "franken.backbones.utils.load_model_registry", return_value=mock_registry
    ), patch(
        "franken.backbones.utils.make_summary", return_value="available backbones"
    ):
        with pytest.raises(NameError) as exc_info:
            franken.backbones.utils.get_checkpoint_path("invalid_backbone")

        assert "Unknown invalid_backbone GNN backbone" in str(exc_info.value)
        assert "available backbones" in str(exc_info.value)


def test_get_checkpoint_path_download_required(mock_registry, mock_cache_folder):
    with patch(
        "franken.backbones.utils.load_model_registry", return_value=mock_registry
    ), patch(
        "franken.backbones.utils.CacheDir.get", return_value=mock_cache_folder
    ), patch("pathlib.Path.exists", return_value=False), patch(
        "franken.backbones.utils.download_checkpoint"
    ) as mock_download:
        result = franken.backbones.utils.get_checkpoint_path("UNIMPLEMENTED_MODEL")
        expected_path = mock_cache_folder / "gnn_checkpoints" / "unimplemented.ckpt"
        assert result == expected_path
        mock_download.assert_called_once_with("UNIMPLEMENTED_MODEL")



import abc
import json
import logging
from pathlib import Path
from typing import Tuple, Union

import torch
import torch.utils.data

from franken.rf.model import FrankenPotential
from franken.rf.scaler import Statistics, compute_dataset_statistics
from franken.trainers.log_utils import (
    DataSplit,
    LogCollection,
    LogEntry,
    dtypeJSONEncoder,
)
from franken.utils.misc import are_dicts_equal


logger = logging.getLogger("franken")


class BaseTrainer(abc.ABC):
    def __init__(
        self,
        train_dataloader: torch.utils.data.DataLoader,
        log_dir: Path | None = None,  # If None, logging is disabled
        save_every_model: bool = True,
        device: Union[torch.device, str, int] = "cpu",
        dtype: Union[str, torch.dtype] = torch.float32,
    ):
        self.log_dir = log_dir
        self.save_every_model = save_every_model
        self.train_dataloader = train_dataloader
        self.statistics_ = None
        if isinstance(dtype, str):
            if dtype.lower() == "float64" or dtype.lower == "double":
                dtype = torch.float64
            elif dtype.lower() == "float32" or dtype.lower == "float":
                dtype = torch.float32
            else:
                raise ValueError(
                    f"Invalid dtype {dtype}. Allowed values are 'float64', 'double', 'float32', 'single'."
                )
        if dtype not in {torch.float32, torch.float64}:
            raise ValueError(
                f"Invalid dtype {dtype}. torch.float32 or torch.float64 are allowed."
            )
        self.buffer_dt = dtype
        self.device = torch.device(device)

    @torch.no_grad()
    def get_statistics(self, model: FrankenPotential) -> Tuple[Statistics, dict]:
        if self.statistics_ is None or not are_dicts_equal(
            self.statistics_[1], model.gnn.init_args()
        ):
            stat = compute_dataset_statistics(
                dataset=self.train_dataloader.dataset,  # type: ignore
                gnn=model.gnn,
                device=self.device,
            )
            stat_dict = model.gnn.init_args()
            self.statistics_ = (stat, stat_dict)

        return self.statistics_

    @abc.abstractmethod
    def on_fit_start(self, model: FrankenPotential):
        pass

    @abc.abstractmethod
    def fit(
        self,
        model: FrankenPotential,
        solver_params: dict,
    ) -> tuple[LogCollection, torch.Tensor]:
        pass

    @abc.abstractmethod
    def evaluate(
        self,
        model: FrankenPotential,
        dataloader: torch.utils.data.DataLoader,
        log_collection: LogCollection,
        all_weights: torch.Tensor,
        metrics: list[str],
    ) -> LogCollection:
        pass

    def serialize_logs(
        self,
        model: FrankenPotential,
        log_collection: LogCollection,
        all_weights: torch.Tensor,
        best_model_split: DataSplit = DataSplit.TRAIN,
    ):
        assert self.log_dir is not None, "Log directory is not set"
        model_hash_set = set(log.checkpoint_hash for log in log_collection)
        assert len(model_hash_set) == 1
        model_hash = model_hash_set.pop()
        log_collection.save_json(self.log_dir / "log.json")

        # Save the model checkpoint
        if self.save_every_model:
            ckpt_dir = self.log_dir / "checkpoints"
            ckpt_dir.mkdir(parents=True, exist_ok=True)
            model_save_path = ckpt_dir / f"{model_hash}.pt"
            model.save(model_save_path, multi_weights=all_weights)
            logger.debug(
                f"Saved multiple models (hash={model_hash}) " f"to {model_save_path}"
            )
        # Log the best model
        self.serialize_best_model(model, all_weights, split=best_model_split)

    def serialize_best_model(
        self,
        model: FrankenPotential,
        all_weights: torch.Tensor,
        split: DataSplit = DataSplit.TRAIN,
    ) -> None:
        assert self.log_dir is not None, "Log directory is not set"
        log_collection = LogCollection.from_json(self.log_dir / "log.json")
        best_model = log_collection.get_best_model(split=split)

        best_model_file = self.log_dir / "best.json"
        should_save = True
        if best_model_file.exists():
            with open(best_model_file, "r") as f:
                current_best = LogEntry.from_dict(json.load(f))
            if best_model == current_best:
                should_save = False

        if should_save:
            logger.debug(f"Identified new best model: {best_model}")
            with open(best_model_file, "w") as f:
                json.dump(best_model.to_dict(), f, indent=4, cls=dtypeJSONEncoder)
            weights = all_weights[best_model.checkpoint_rf_weight_id]
            model.rf.weights = weights.reshape_as(model.rf.weights)
            model.save(self.log_dir / "best_ckpt.pt")
            logger.debug(
                f"Saved best model (within-experiment ID={best_model.checkpoint_rf_weight_id}) "
                f"to {self.log_dir / 'best_ckpt.pt'}"
            )



from typing import Mapping

import torch


class AtomicEnergiesShift(torch.nn.Module):
    atomic_energies: torch.Tensor
    Z_keys: list[int]

    def __init__(
        self,
        num_species: int,
        atomic_energies: Mapping[int, torch.Tensor | float] | None = None,
    ):
        """
        Initialize the AtomicEnergiesShift module.

        Args:
            num_species:
            atomic_energies: A dictionary mapping atomic numbers to atomic energies.
        """
        super().__init__()

        self.num_species = num_species
        self.register_buffer("atomic_energies", torch.zeros(num_species))
        self.Z_keys = []
        self.is_initialized = False

        if atomic_energies is not None:
            self.set_from_atomic_energies(atomic_energies)

    def set_from_atomic_energies(
        self, atomic_energies: Mapping[int, torch.Tensor | float]
    ):
        assert (
            len(atomic_energies) == self.num_species
        ), f"{len(atomic_energies)=} != {self.num_species=}"
        device = self.atomic_energies.device
        self.atomic_energies = torch.stack(
            [
                v.clone().detach() if isinstance(v, torch.Tensor) else torch.tensor(v)
                for v in atomic_energies.values()
            ]
        ).to(device)
        self.Z_keys = list(atomic_energies.keys())
        self.is_initialized = True

    def get_extra_state(self):
        return {"Z_keys": torch.tensor(self.Z_keys, dtype=torch.int32)}

    def set_extra_state(self, state):
        self.Z_keys = state["Z_keys"].tolist()

    def forward(self, atomic_numbers: torch.Tensor) -> torch.Tensor:
        """
        Calculate the energy shift for a given set of atomic numbers.

        Args:
            atomic_numbers: A tensor containing atomic numbers for which to calculate the energy shift.

        Returns:
            A tensor representing the total energy shift for the provided atomic numbers.
        """

        shift = torch.tensor(
            0.0, dtype=self.atomic_energies.dtype, device=self.atomic_energies.device
        )

        for z, atom_ene in zip(self.Z_keys, self.atomic_energies):
            mask = z == atomic_numbers
            shift += torch.sum(atom_ene * mask)

        return shift

    def __repr__(self):
        formatted_energies = " , ".join(
            [
                f"{z}: {atom_ene}"
                for z, atom_ene in zip(self.Z_keys, self.atomic_energies)
            ]
        )
        return f"{self.__class__.__name__}({formatted_energies})"



import math
from typing import Any, Optional

import torch
import torch.utils.data
from tqdm import tqdm

import franken.utils.distributed as dist_utils
from franken.data import BaseAtomsDataset
from franken.utils.misc import garbage_collection_cuda


class Statistics:
    """Running statistics on a dataset, computing both per-species and global statistics."""

    def __init__(self, input_dim: Optional[int] = None):
        self.input_dim = input_dim
        self.statistics: dict[int, dict[str, Any]] = {}
        if self.input_dim is not None:
            self._initialize_statistics_for_prefix(0)  # Use 0 for global statistics

    def __call__(self, descriptors: torch.Tensor, atomic_numbers: torch.Tensor) -> None:
        """Call self.update(descriptors, atomic_numbers)"""
        self.update(descriptors, atomic_numbers)

    def update(self, descriptors: torch.Tensor, atomic_numbers: torch.Tensor) -> None:
        """
        Update running statistics for both per-species and global features.

        Args:
            descriptors: Input features of shape (n_atoms, input_dim)
            atomic_numbers: Atomic numbers for each atom
        """
        if self.input_dim is None:
            self.input_dim = descriptors.shape[1]
        assert self.input_dim == descriptors.shape[1]

        # Update global statistics (use 0 as the key)
        self._update_stats(descriptors, 0)

        # Update per-species statistics directly using atomic numbers as integer keys
        for Z in torch.unique(atomic_numbers, sorted=True):
            mask = atomic_numbers == Z
            self._update_stats(descriptors[mask], Z.item())  # Use Z as an integer key

    def _update_stats(self, x: torch.Tensor, prefix: int) -> None:
        """Update running statistics using Welford's online algorithm."""
        assert (
            x.ndim == 2 and x.shape[1] == self.input_dim
        ), f"Expected shape (n, {self.input_dim})"

        # Initialize statistics if they don't exist
        if prefix not in self.statistics:
            self._initialize_statistics_for_prefix(prefix, device=x.device)

        stats = self.statistics[prefix]

        # Compute batch statistics
        batch_size = x.shape[0]
        batch_mean = torch.mean(x, dim=0)
        batch_m2 = torch.sum((x - batch_mean) ** 2, dim=0)

        # Update running statistics (Welford's algorithm)
        new_count = stats["count"] + batch_size
        delta = batch_mean - stats["mean"]
        new_mean = stats["mean"] + delta * batch_size / new_count
        new_M2 = (
            stats["M2"]
            + batch_m2
            + delta**2 * (batch_size * stats["count"] / new_count)
        )

        # Store updated statistics
        stats["count"] = new_count
        stats["mean"] = new_mean
        stats["M2"] = new_M2
        stats["std"] = torch.sqrt(new_M2 / new_count)
        stats["min"] = torch.minimum(stats["min"], torch.min(x, dim=0).values)
        stats["max"] = torch.maximum(stats["max"], torch.max(x, dim=0).values)

    def _initialize_statistics_for_prefix(
        self, prefix: int, device: torch.device = torch.device("cpu")
    ) -> None:
        """Initialize the statistics dictionary for a given prefix if it does not exist."""
        assert self.input_dim is not None
        self.statistics[prefix] = {
            "count": 0,
            "mean": torch.zeros(self.input_dim, device=device, dtype=torch.float64),
            "M2": torch.zeros(self.input_dim, device=device, dtype=torch.float64),
            "std": torch.ones(self.input_dim, device=device, dtype=torch.float64),
            "min": torch.full(
                (self.input_dim,), float("inf"), device=device, dtype=torch.float64
            ),
            "max": torch.full(
                (self.input_dim,), float("-inf"), device=device, dtype=torch.float64
            ),
        }

    def __repr__(self):
        return f"{self.__class__.__name__}({self.statistics.__repr__()})"

    @property
    def num_species(self):
        return len(self.statistics.keys()) - 1


class FeatureScaler(torch.nn.Module):
    def __init__(
        self,
        input_dim: int,
        statistics: Optional[Statistics],
        scale_by_Z: bool,
        num_species: int,
    ):
        super().__init__()
        self.input_dim = input_dim
        self.scale_by_Z = scale_by_Z
        self.num_species = num_species if scale_by_Z else 1
        self.Z_keys: list[int] = []

        # Initialize mean and std to default values
        self.mean: torch.Tensor
        self.std: torch.Tensor
        self.register_buffer("mean", torch.zeros(self.num_species, input_dim))
        self.register_buffer("std", torch.ones(self.num_species, input_dim))

        if statistics is not None:
            self.set_from_statistics(statistics)

    def set_from_statistics(self, statistics: Statistics):
        assert self.input_dim == statistics.input_dim
        assert self.input_dim is not None

        if self.scale_by_Z:
            # Initialize Z_keys as a sorted list of atomic numbers (excluding global 0)
            self.Z_keys = sorted(
                [key for key in statistics.statistics.keys() if key != 0]
            )
            assert len(self.Z_keys) == self.num_species

            for idx, Z in enumerate(self.Z_keys):
                self.mean[idx] = statistics.statistics[Z]["mean"].to(
                    dtype=self.mean.dtype
                )
                self.std[idx] = statistics.statistics[Z]["std"].to(dtype=self.std.dtype)
        else:
            assert self.num_species == 1
            # Use global scaling: keep mean and std as [1, input_dim]
            global_stats = statistics.statistics[0]  # Use key 0 for global
            self.mean = global_stats["mean"].view(1, -1).to(dtype=self.mean.dtype)
            self.std = global_stats["std"].view(1, -1).to(dtype=self.std.dtype)

    def forward(
        self, descriptors: torch.Tensor, atomic_numbers: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        assert descriptors.shape[1] == self.input_dim

        if not self.scale_by_Z:
            # Global scaling, simply normalize all descriptors
            return (descriptors - self.mean) / (self.std * math.sqrt(self.input_dim))

        # Normalize per atomic number (Z)
        assert atomic_numbers is not None, "atomic_numbers required when scaling by Z"
        for idx, Z in enumerate(self.Z_keys):
            mask = atomic_numbers == Z
            descriptors[mask] = (descriptors[mask] - self.mean[idx]) / (
                self.std[idx] * math.sqrt(self.input_dim)
            )
        return descriptors

    def hyperparameters(self):
        return self.init_args()

    def init_args(self):
        return {"scale_by_Z": self.scale_by_Z}

    def get_extra_state(self):
        return {"Z_keys": torch.tensor(self.Z_keys, dtype=torch.int32)}

    def set_extra_state(self, state):
        self.Z_keys = state["Z_keys"].tolist()

    def __repr__(self):
        repr = f"scale_by_Z={self.scale_by_Z}, Z_keys={self.Z_keys}, mean={self.mean.shape}, std={self.std.shape}"
        return f"{self.__class__.__name__}({repr})"


def compute_dataset_statistics(
    dataset: BaseAtomsDataset,
    gnn,
    device: torch.device,
):
    # Initialize statistics object
    gnn_features_stats = Statistics()
    pbar = tqdm(
        total=len(dataset),
        desc="Computing dataset statistics",
        disable=dist_utils.get_rank() != 0,
    )
    # Iterate over dataset and update stats
    for data, _ in dataset:  # type: ignore
        data = data.to(device=device)
        gnn_features = gnn.descriptors(data)
        gnn_features_stats.update(gnn_features, atomic_numbers=data.atomic_numbers)
        pbar.update(1)

    # Cleanup
    garbage_collection_cuda()

    return gnn_features_stats



from typing import Literal


elements = [
    "Ag",
    "Au",
    "Cd",
    "Co",
    "Cr",
    "Cu",
    "Fe",
    "Hf",
    "Hg",
    "Ir",
    "Mn",
    "Mo",
    "Nb",
    "Ni",
    "Os",
    "Pd",
    "Pt",
    "Re",
    "Rh",
    "Ru",
    "Ta",
    "Tc",
    "Ti",
    "V",
    "W",
    "Zn",
    "Zr",
]
MELTING_POINTS = {
    "Ag": 1235,
    "Au": 1337,
    "Cd": 594,
    "Co": 1768,
    "Cr": 2180,
    "Cu": 1358,
    "Fe": 1811,
    "Hf": 2506,
    "Hg": 234,
    "Ir": 2739,
    "Mn": 1519,
    "Mo": 2896,
    "Nb": 2750,
    "Ni": 1728,
    "Os": 3306,
    "Pd": 1828,
    "Pt": 2041,
    "Re": 3459,
    "Rh": 2237,
    "Ru": 2607,
    "Ta": 3290,
    "Tc": 2430,
    "Ti": 1941,
    "V": 2183,
    "W": 3695,
    "Zn": 693,
    "Zr": 2128,
}
TEMPERATURE_LABELS = ["cold", "warm", "melt"]
TEMPERATURE_MULTIPLIERS = {
    "cold": 0.25,
    "warm": 0.75,
    "melt": 1.25,
}


def tm23_temp(
    element: str, temperature_label: Literal["cold", "warm", "melt"]
) -> float:
    if temperature_label not in TEMPERATURE_LABELS:
        raise ValueError(
            f"'temperature_label' must be one of {TEMPERATURE_LABELS} "
            f"but found '{temperature_label}'."
        )
    if element not in elements:
        raise ValueError(f"Element '{element}' is not a valid TM23 element label.")
    melt_point = MELTING_POINTS[element]
    sim_temp = melt_point * TEMPERATURE_MULTIPLIERS[temperature_label]
    return sim_temp



import torch

from franken.metrics.base import BaseMetric


class MetricRegistry:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._metrics = {}
        return cls._instance

    def register(self, name: str, metric_class: type) -> None:
        """Register a metric class"""
        self._metrics[name] = metric_class

    def init_metric(
        self, name: str, device: torch.device, dtype: torch.dtype = torch.float32
    ) -> BaseMetric:
        """Create a new instance of a metric"""
        if name not in self._metrics:
            raise KeyError(
                f"Metric '{name}' not found. Available metrics: {list(self._metrics.keys())}"
            )
        return self._metrics[name](device=device, dtype=dtype)

    @property
    def available_metrics(self) -> list[str]:
        return list(self._metrics.keys())


registry = MetricRegistry()



from pathlib import Path
from typing import Optional

import ase
import numpy as np
import sevenn._keys as KEY
import torch
from sevenn.atom_graph_data import AtomGraphData
from sevenn.train.dataload import atoms_to_graph
from tqdm import tqdm

import franken.utils.distributed as dist_utils
from franken.backbones import REGISTRY
from franken.backbones.utils import get_checkpoint_path
from franken.data import BaseAtomsDataset, Configuration, Target


class SevennAtomsToGraphs:
    """
    Args:
        cutoff (float): Cutoff in Angstrom to build atoms graph from positions
        transfer_info (bool): if True, copy info from atoms to graph
    """

    def __init__(self, cutoff: float, transfer_info: bool, y_from_calc: bool):
        self.cutoff = cutoff
        self.transfer_info = transfer_info
        self.y_from_calc = y_from_calc

    def convert(self, atoms: ase.Atoms):
        if not self.y_from_calc:
            # It means we're not interested in forces and energies.
            # workaround is to set the attributes to invalid and then
            # remove the attributes
            atoms.info["y_energy"] = np.nan
            atoms.arrays["y_force"] = np.full(atoms.arrays["positions"].shape, np.nan)
        graph = atoms_to_graph(
            atoms,
            cutoff=self.cutoff,
            transfer_info=self.transfer_info,
            y_from_calc=self.y_from_calc,
        )
        if not self.y_from_calc:
            del graph[KEY.ENERGY]
            del graph[KEY.FORCE]
            del atoms.info["y_energy"]
            del atoms.arrays["y_force"]
        atom_graph_data = AtomGraphData.from_numpy_dict(graph)
        return atom_graph_data

    def convert_all(
        self,
        atoms_list,
        process_rank: Optional[int] = None,
        split_name: Optional[str] = None,
    ):
        graphs = []
        atoms_iter = atoms_list
        if process_rank is None:
            process_rank = dist_utils.get_rank()
        if process_rank == 0:
            desc = "ASE -> SEVENN"
            if split_name is not None:
                desc += f" ({split_name})"
            atoms_iter = tqdm(atoms_list, desc=desc)
        for atoms in atoms_iter:
            graphs.append(self.convert(atoms))
        return graphs


class SevennAtomsDataset(BaseAtomsDataset):
    def __init__(
        self,
        data_path: str | Path | None,
        split: str,
        num_random_subsamples: int | None = None,
        subsample_rng: int | None = None,
        gnn_backbone_id: str | None = None,
        cutoff: float = 6.0,
        precompute=True,
    ):
        super().__init__(data_path, split, num_random_subsamples, subsample_rng)
        if gnn_backbone_id is not None:
            cutoff = self.load_info_from_gnn_config(gnn_backbone_id)
        else:
            assert cutoff is not None

        if split == "md":
            self.a2g = SevennAtomsToGraphs(
                cutoff, transfer_info=False, y_from_calc=False
            )
        else:
            self.a2g = SevennAtomsToGraphs(
                cutoff, transfer_info=False, y_from_calc=True
            )
        self.graphs = None
        if precompute and len(self.ase_atoms) > 0:
            self.graphs = self.a2g.convert_all(
                self.ase_atoms,
                split_name=self.split,
            )

    def load_info_from_gnn_config(self, gnn_backbone_id: str):
        ckpt_path = get_checkpoint_path(gnn_backbone_id)
        backbone_family = REGISTRY[gnn_backbone_id]["kind"].lower()
        if backbone_family != "sevenn":
            raise ValueError(f"Incompatible backbone family {backbone_family}")
        checkpoint = torch.load(ckpt_path, map_location="cpu", weights_only=False)
        config = checkpoint["config"]
        cutoff = config["cutoff"]
        del checkpoint, config
        return cutoff

    def __getitem__(self, idx, no_targets: bool = False):
        """Returns an array of (inputs, outputs) with inputs being a configuration
        and outputs being the target (energy and forces).
        Note: ONLY for the 'train' split, the energy_shift is removed from the target.
        """
        if self.graphs is None:
            graph = self.a2g.convert(self.ase_atoms[idx])
        else:
            graph = self.graphs[idx]

        data = Configuration(
            atom_pos=graph.pos,
            atomic_numbers=graph[KEY.ATOMIC_NUMBERS],
            natoms=torch.tensor(len(graph[KEY.ATOMIC_NUMBERS])).view(1),
            edge_index=graph.edge_index,
            shifts=graph[KEY.CELL_SHIFT],
            cell=graph[KEY.CELL],
            batch_ids=(
                graph.batch
                if graph.batch is not None
                else torch.zeros(graph[KEY.NUM_ATOMS], dtype=torch.int64)
            ),
        )
        if no_targets:
            return data

        energy = graph[KEY.ENERGY]
        if self.split == "train":
            energy = energy - self.energy_shifts[idx]
        target = Target(
            energy=energy,
            forces=graph[KEY.FORCE],
        )
        return data, target



import numpy as np
from franken.utils.linalg.cov import (
    _lowmemcov_rank1_update,
    _lowmemcov_rankk_update,
    _rank1_update,
    _rankk_update,
    rank1_update,
    rankk_update
)
from franken.utils.linalg.psdsolve import (
    _lowmem_psd_ridge,
    _naive_psd_ridge,
    psd_ridge
)
from franken.utils.linalg.tri import (
    _trilerp_cpu,
    _trilerp_triton,
    inplace_triangular_divide,
    triangular_lerp
)
from .conftest import (
    DEVICES,
    SKIP_NO_CUDA,
)

import pytest
import torch

from functools import partial
from unittest.mock import DEFAULT, patch


class TestCovUpdates:
    mat_size = 10

    def A(self, device, dtype):
        gen = torch.Generator(device=device)
        gen.manual_seed(37)
        mat = torch.randn(self.mat_size, self.mat_size, device=device, dtype=dtype, generator=gen)
        return (mat @ mat.T) / self.mat_size

    def d(self, device, dtype):
        gen = torch.Generator(device=device)
        gen.manual_seed(38)
        diag = torch.randn(self.mat_size, device=device, dtype=dtype, generator=gen)
        return diag

    def vec(self, device, dtype):
        gen = torch.Generator(device=device)
        gen.manual_seed(39)
        vec = torch.randn(self.mat_size, device=device, dtype=dtype, generator=gen)
        return vec

    def veck(self, device, dtype):
        gen = torch.Generator(device=device)
        gen.manual_seed(40)
        veck = torch.randn((self.mat_size, 3), device=device, dtype=dtype, generator=gen)
        return veck

    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
    @pytest.mark.parametrize("upper", [True, False])
    def test_lowmem_rank1(self, dtype, device, upper):
        A = self.A(device, dtype)
        Acp = torch.clone(A)
        d = self.d(device, dtype)
        vec = self.vec(device, dtype)

        tri_func = partial(torch.triu, diagonal=1) if upper else partial(torch.tril, diagonal=-1)
        other_tri_func = partial(torch.tril, diagonal=-1) if upper else partial(torch.triu, diagonal=1)
        exp_cov = A + tri_func(torch.outer(vec, vec))
        exp_diag = d + vec ** 2

        # TEST FUNCTION HERE
        out_cov, out_diag = _lowmemcov_rank1_update(A, d, vec, upper=upper)

        # must be in-place
        assert out_cov.data_ptr() == A.data_ptr()
        assert out_diag.data_ptr() == d.data_ptr()
        # correctness
        torch.testing.assert_close(tri_func(out_cov), tri_func(exp_cov))
        torch.testing.assert_close(out_diag, exp_diag)
        # preserving the other triangular matrix
        torch.testing.assert_close(other_tri_func(out_cov), other_tri_func(Acp))

    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
    @pytest.mark.parametrize("upper", [True, False])
    def test_lowmem_rankk(self, device, dtype, upper):
        A = self.A(device, dtype)
        Acp = torch.clone(A)
        d = self.d(device, dtype)
        veck = self.veck(device, dtype)

        tri_func = partial(torch.triu, diagonal=1) if upper else partial(torch.tril, diagonal=-1)
        other_tri_func = partial(torch.tril, diagonal=-1) if upper else partial(torch.triu, diagonal=1)
        exp_cov = A + tri_func(veck @ veck.T)
        exp_diag = d + (veck ** 2).sum(1)

        # TEST FUNCTION HERE
        out_cov, out_diag = _lowmemcov_rankk_update(A, d, veck, upper=upper)

        # must be in-place
        assert out_cov.data_ptr() == A.data_ptr()
        assert out_diag.data_ptr() == d.data_ptr()
        # correctness
        torch.testing.assert_close(tri_func(out_cov), tri_func(exp_cov))
        torch.testing.assert_close(out_diag, exp_diag)
        # preserving the other triangular matrix
        torch.testing.assert_close(other_tri_func(out_cov), other_tri_func(Acp))

    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
    @pytest.mark.parametrize("upper", [True, False])
    def test_nocupy_rank1(self, dtype, device, upper):
        A = self.A(device, dtype)
        Acp = torch.clone(A)
        d = self.d(device, dtype)
        vec = self.vec(device, dtype)

        tri_func = partial(torch.triu, diagonal=1) if upper else partial(torch.tril, diagonal=-1)
        other_tri_func = partial(torch.tril, diagonal=-1) if upper else partial(torch.triu, diagonal=1)
        exp_cov = A + tri_func(torch.outer(vec, vec))
        exp_diag = d + vec ** 2

        with patch.multiple(
            "franken.utils.linalg.cov", cupy=None
        ) as _:
            # TEST FUNCTION HERE
            out_cov, out_diag = _rank1_update(A, d, vec, upper=upper)
        # must be in-place
        assert out_cov.data_ptr() == A.data_ptr()
        assert out_diag.data_ptr() == d.data_ptr()
        # correctness
        torch.testing.assert_close(tri_func(out_cov), tri_func(exp_cov))
        torch.testing.assert_close(out_diag, exp_diag)
        # preserving the other triangular matrix
        torch.testing.assert_close(other_tri_func(out_cov), other_tri_func(Acp))

    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
    @pytest.mark.parametrize("upper", [True, False])
    def test_nocupy_rankk(self, device, dtype, upper):
        A = self.A(device, dtype)
        Acp = torch.clone(A)
        d = self.d(device, dtype)
        veck = self.veck(device, dtype)

        tri_func = partial(torch.triu, diagonal=1) if upper else partial(torch.tril, diagonal=-1)
        other_tri_func = partial(torch.tril, diagonal=-1) if upper else partial(torch.triu, diagonal=1)
        exp_cov = A + tri_func(veck @ veck.T)
        exp_diag = d + (veck ** 2).sum(1)

        with patch.multiple(
            "franken.utils.linalg.cov", cupy=None
        ) as _:
            # TEST FUNCTION HERE
            out_cov, out_diag = _rankk_update(A, d, veck, upper=upper)
        # must be in-place
        assert out_cov.data_ptr() == A.data_ptr()
        assert out_diag.data_ptr() == d.data_ptr()
        # correctness
        torch.testing.assert_close(tri_func(out_cov), tri_func(exp_cov))
        torch.testing.assert_close(out_diag, exp_diag)
        # preserving the other triangular matrix
        torch.testing.assert_close(other_tri_func(out_cov), other_tri_func(Acp))

    @SKIP_NO_CUDA
    def test_lowmem_cuda_nocupy(self):
        upper = True
        dtype = torch.float32
        A = self.A("cuda", dtype)
        d = self.d("cuda", dtype)
        vec = self.vec("cuda", dtype)
        veck = self.veck("cuda", dtype)

        with patch.multiple(
            "franken.utils.linalg.cov", cupy=None
        ) as _:
            with pytest.raises(AssertionError):
                _lowmemcov_rank1_update(A, d, vec, upper=upper)
            with pytest.raises(AssertionError):
                _lowmemcov_rankk_update(A, d, veck, upper=upper)

    @SKIP_NO_CUDA
    def test_rank1_dispatcher_cuda(self):
        upper = True
        dtype = torch.float32
        A = self.A("cuda", dtype)
        d = self.d("cuda", dtype)
        vec = self.vec("cuda", dtype)
        with patch.multiple("franken.utils.linalg.cov", _lowmemcov_rank1_update=DEFAULT, _rank1_update=DEFAULT) as mocks:
            with patch.multiple("franken.utils.linalg.cov", cupy=None):
                with pytest.warns(UserWarning, match="`cupy` is not available"):
                    rank1_update(A, d, vec, upper=upper)
            mocks['_rank1_update'].assert_called_once()
            rank1_update(A, d, vec, upper=upper)
            mocks['_lowmemcov_rank1_update'].assert_called_once()

    @SKIP_NO_CUDA
    def test_rankk_dispatcher_cuda(self):
        upper = True
        dtype = torch.float32
        A = self.A("cuda", dtype)
        d = self.d("cuda", dtype)
        veck = self.veck("cuda", dtype)
        with patch.multiple("franken.utils.linalg.cov", _lowmemcov_rankk_update=DEFAULT, _rankk_update=DEFAULT) as mocks:
            with patch.multiple("franken.utils.linalg.cov", cupy=None):
                with pytest.warns(UserWarning, match="`cupy` is not available"):
                    rankk_update(A, d, veck, upper=upper)
            mocks['_rankk_update'].assert_called_once()
            rankk_update(A, d, veck, upper=upper)
            mocks['_lowmemcov_rankk_update'].assert_called_once()


class TestPSDSolvers:
    mat_size = 5  # needs to be small otherwise too much numerical error

    def A(self, device, dtype):
        gen = torch.Generator(device=device)
        gen.manual_seed(37)
        mat = torch.randn(self.mat_size, self.mat_size, device=device, dtype=dtype, generator=gen)
        return (mat @ mat.T) / self.mat_size

    def B(self, device, dtype):
        gen = torch.Generator(device=device)
        gen.manual_seed(33)
        return torch.randn(self.mat_size, 2, device=device, dtype=dtype, generator=gen) / self.mat_size

    def expected(self, A, B, penalty):
        A = A + torch.eye(A.shape[0], dtype=A.dtype, device=A.device) * penalty
        return torch.linalg.solve(A.double(), B.double()).to(dtype=A.dtype)

    @SKIP_NO_CUDA
    @pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
    def test_lowmem_cuda(self, dtype):
        A = self.A("cuda", dtype)
        B = self.B("cuda", dtype)
        penalty = 1e-5
        expected = self.expected(A, B, penalty)

        # To make sure only upper part of A is accessed set the lower to zero
        A = torch.triu(A)
        Acopy = torch.clone(A)
        result = _lowmem_psd_ridge(A, B, penalty)

        # A should have been overwritten (by cholesky)
        assert not torch.isclose(A, Acopy).all()
        # rhs MAY be overwritten by result
        # (only true if B is F-contiguous which is not the case here.)
        #assert result.data_ptr() == B.data_ptr()
        # correctness
        torch.testing.assert_close(result, expected)

    def test_lowmem_cpu(self):
        A = self.A("cpu", torch.float32)
        B = self.B("cpu", torch.float32)
        penalty = 1e-5
        with pytest.raises(AssertionError):
            _lowmem_psd_ridge(A, B, penalty)

    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
    def test_naive(self, device, dtype):
        A = self.A(device, dtype)
        B = self.B(device, dtype)
        penalty = 1e-5
        expected = self.expected(A, B, penalty)

        # To make sure only upper part of A is accessed set the lower to zero
        A = torch.triu(A)
        Acopy = torch.clone(A)
        Bcopy = torch.clone(B)
        result = _naive_psd_ridge(A, B, penalty)

        # A should have been been overwritten (only diagonal)
        assert not torch.isclose(A, Acopy, rtol=penalty / 10, atol=penalty / 10).all()
        # rhs should be NOT overwritten by result
        torch.testing.assert_close(B, Bcopy)
        # correctness
        torch.testing.assert_close(result, expected)

    @SKIP_NO_CUDA
    def test_dispatcher_cuda(self):
        with patch.multiple(
            "franken.utils.linalg.psdsolve", _naive_psd_ridge=DEFAULT, _lowmem_psd_ridge=DEFAULT
        ) as mocks:
            A = self.A("cuda", torch.float32)
            B = self.B("cuda", torch.float32)
            psd_ridge(A, B, 1e-5)
            mocks['_lowmem_psd_ridge'].assert_called_once()

    @SKIP_NO_CUDA
    def test_dispatcher_cuda_nocupy(self):
        with patch.multiple(
            "franken.utils.linalg.psdsolve", _naive_psd_ridge=DEFAULT, _lowmem_psd_ridge=DEFAULT, cupy=None
        ) as mocks:
            A = self.A("cuda", torch.float32)
            B = self.B("cuda", torch.float32)
            with pytest.warns(UserWarning, match="`cupy` is not available"):
                psd_ridge(A, B, 1e-5)
            mocks['_naive_psd_ridge'].assert_called_once()

    def test_dispatcher_cpu(self):
        with patch.multiple(
            "franken.utils.linalg.psdsolve", _naive_psd_ridge=DEFAULT, _lowmem_psd_ridge=DEFAULT
        ) as mocks:
            A = self.A("cpu", torch.float32)
            B = self.B("cpu", torch.float32)
            psd_ridge(A, B, 1e-5)
            mocks['_naive_psd_ridge'].assert_called_once()


@pytest.mark.parametrize("device", DEVICES)
class TestTridivide:
    def test_tridivide_up(self, device):
        mat = np.random.randn(5, 5)
        val = 12.0
        triu_ids = np.triu_indices(mat.shape[0], 1)
        expected = np.copy(mat)
        expected[triu_ids] /= val

        expected_torch = torch.from_numpy(expected).to(device=device)
        mat_torch = torch.from_numpy(mat).to(device=device)
        actual = inplace_triangular_divide(mat_torch, val, upper=True)
        torch.testing.assert_close(actual, expected_torch)

    def test_tridivide_lo(self, device):
        mat = np.random.randn(5, 5)
        val = 12.0
        tril_ids = np.tril_indices(mat.shape[0], -1)
        expected = np.copy(mat)
        expected[tril_ids] /= val

        expected_torch = torch.from_numpy(expected).to(device=device)
        mat_torch = torch.from_numpy(mat).to(device=device)
        actual = inplace_triangular_divide(mat_torch, val, upper=False)
        torch.testing.assert_close(actual, expected_torch)


class TestTrilerp:
    @pytest.mark.parametrize("weight", [0.0, 0.2, 1.0])
    def test_cpu_noinplace(self, weight):
        mat = torch.randn(5, 5)
        triu = torch.triu(mat, diagonal=1)
        tril = torch.tril(mat, diagonal=-1)
        expected = triu + weight * (tril.T - triu)

        actual = _trilerp_cpu(mat, weight, None)
        torch.testing.assert_close(torch.triu(actual, diagonal=1), expected)
        assert mat.data_ptr() != actual.data_ptr(), "This should not be inplace"

    @pytest.mark.parametrize("weight", [0.0, 0.2, 1.0])
    def test_cpu_inplace(self, weight):
        mat = torch.randn(5, 5)
        triu = torch.triu(mat, diagonal=1)
        tril = torch.tril(mat, diagonal=-1)
        expected = triu + weight * (tril.T - triu)

        actual = _trilerp_cpu(mat, weight, out=mat)
        assert mat.data_ptr() == actual.data_ptr(), "This should be inplace"
        torch.testing.assert_close(torch.triu(actual, diagonal=1), expected)

    @SKIP_NO_CUDA
    @pytest.mark.parametrize("weight", [0.0, 0.2, 1.0])
    def test_cuda_noinplace(self, weight):
        mat = torch.randn(5, 5, device="cuda")
        triu = torch.triu(mat, diagonal=1)
        tril = torch.tril(mat, diagonal=-1)
        expected = triu + weight * (tril.T - triu)

        actual = _trilerp_triton(mat, weight, None)
        torch.testing.assert_close(torch.triu(actual, diagonal=1), expected)
        assert mat.data_ptr() != actual.data_ptr(), "This should not be inplace"

    @SKIP_NO_CUDA
    @pytest.mark.parametrize("weight", [0.0, 0.2, 1.0])
    def test_cuda_inplace(self, weight):
        mat = torch.randn(5, 5, device="cuda")
        weight = 0.2
        triu = torch.triu(mat, diagonal=1)
        tril = torch.tril(mat, diagonal=-1)
        expected = triu + weight * (tril.T - triu)

        actual = _trilerp_triton(mat, weight, out=mat)
        assert mat.data_ptr() == actual.data_ptr(), "This should be inplace"
        torch.testing.assert_close(torch.triu(actual, diagonal=1), expected)

    @pytest.mark.parametrize("device", DEVICES)
    def test_wrapper(self, device):
        mat = torch.randn(5, 5, device=device)
        weight = 0.2
        triu = torch.triu(mat, diagonal=1)
        tril = torch.tril(mat, diagonal=-1)
        expected_mat = triu + weight * (tril.T - triu)
        diag_up = torch.randn(5, device=device)
        diag_lo = torch.randn(5, device=device)
        expected_diag = diag_up + weight * (diag_lo - diag_up)

        actual_mat, actual_diag = triangular_lerp(mat, diag_up, diag_lo, weight, inplace=True)
        assert mat.data_ptr() == actual_mat.data_ptr(), "This should be inplace"
        assert actual_diag.data_ptr() == diag_up.data_ptr(), "This should be inplace"
        torch.testing.assert_close(torch.triu(actual_mat, diagonal=1), expected_mat)
        torch.testing.assert_close(actual_diag, expected_diag)


from functools import wraps
from typing import Any, Callable, Optional, Sequence, Tuple, Union

import torch
from torch._functorch.eager_transforms import (
    _construct_standard_basis_for,
    _jvp_with_argnums,
    _slice_argnums,
    error_if_complex,
    safe_unflatten,
)
from torch._functorch.utils import argnums_t
from torch.func import vmap
from torch.utils._pytree import tree_flatten, tree_unflatten

from franken.utils.misc import garbage_collection_cuda, is_cuda_out_of_memory


def jacfwd(
    # drop-in replacement of torch.func.jacfwd accepting the chunk_size argument (as with jacrev)
    func: Callable,
    argnums: argnums_t = 0,
    has_aux: bool = False,
    *,
    randomness: str = "error",
    chunk_size: Optional[int] = None,
):
    def wrapper_fn(*args):
        error_if_complex("jacfwd", args, is_input=True)
        primals = args if argnums is None else _slice_argnums(args, argnums)
        flat_primals, primals_spec = tree_flatten(primals)
        flat_primals_numels = tuple(p.numel() for p in flat_primals)
        flat_basis = _construct_standard_basis_for(flat_primals, flat_primals_numels)
        basis = tree_unflatten(flat_basis, primals_spec)

        def push_jvp(basis):
            output = _jvp_with_argnums(
                func, args, basis, argnums=argnums, has_aux=has_aux
            )
            # output[0] is the output of `func(*args)`
            error_if_complex("jacfwd", output[0], is_input=False)
            if has_aux:
                _, jvp_out, aux = output
                return jvp_out, aux
            _, jvp_out = output
            return jvp_out

        results = vmap(push_jvp, randomness=randomness, chunk_size=chunk_size)(basis)
        if has_aux:
            results, aux = results
            # aux is in the standard basis format, e.g. NxN matrix
            # We need to fetch the first element as original `func` output
            flat_aux, aux_spec = tree_flatten(aux)
            flat_aux = [value[0] for value in flat_aux]
            aux = tree_unflatten(flat_aux, aux_spec)

        jac_outs, spec = tree_flatten(results)
        # Most probably below output check can never raise an error
        # as jvp should test the output before
        # assert_non_empty_output(jac_outs, 'jacfwd(f, ...)(*args)')

        jac_outs_ins = tuple(
            tuple(
                safe_unflatten(jac_out_in, -1, primal.shape)
                for primal, jac_out_in in zip(
                    flat_primals,
                    jac_out.movedim(0, -1).split(flat_primals_numels, dim=-1),
                )
            )
            for jac_out in jac_outs
        )
        jac_outs_ins = tuple(
            tree_unflatten(jac_ins, primals_spec) for jac_ins in jac_outs_ins
        )

        if isinstance(argnums, int):
            jac_outs_ins = tuple(jac_ins[0] for jac_ins in jac_outs_ins)
        if has_aux:
            return tree_unflatten(jac_outs_ins, spec), aux
        return tree_unflatten(jac_outs_ins, spec)

    # Dynamo does not support HOP composition if their inner function is
    # annotated with @functools.wraps(...). We circumvent this issue by applying
    # wraps only if we're not tracing with dynamo.
    if not torch._dynamo.is_compiling():
        wrapper_fn = wraps(func)(wrapper_fn)

    return wrapper_fn


def tune_jacfwd_chunksize(
    test_sample: Sequence[Union[torch.Tensor, Any]],
    mode: str = "power",
    init_val: int = 32,
    max_trials: int = 25,
    **jac_kwargs,
):
    try:
        # We want to tune this and set it ourselves.
        jac_kwargs.pop("chunk_size")
    except KeyError:
        pass

    # Initially we just double in size until an OOM is encountered
    new_size, _ = _adjust_batch_size(
        test_sample, init_val, value=init_val, **jac_kwargs
    )  # initially set to init_val
    if mode == "power":
        new_size = _run_power_scaling(new_size, max_trials, test_sample, **jac_kwargs)
    else:
        raise ValueError("mode in method `scale_batch_size` can only be `power`")

    garbage_collection_cuda()
    return new_size


def _run_power_scaling(new_size, max_trials, test_sample, **jac_kwargs) -> int:
    """Batch scaling mode where the size is doubled at each iteration until an
    OOM error is encountered."""
    for _ in range(max_trials):
        garbage_collection_cuda()
        try:
            # Try jacfwd
            for _ in range(1):
                jacfwd(**jac_kwargs, chunk_size=new_size)(*test_sample)
            # Double in size
            new_size, changed = _adjust_batch_size(
                test_sample, new_size, factor=2.0, **jac_kwargs
            )
        except RuntimeError as exception:
            # Only these errors should trigger an adjustment
            if is_cuda_out_of_memory(exception):
                # If we fail in power mode, half the size and return
                garbage_collection_cuda()
                new_size, _ = _adjust_batch_size(
                    test_sample, new_size, factor=0.5, **jac_kwargs
                )
                break
            else:
                raise  # some other error not memory related
        if not changed:
            # No change in batch size, so we can exit.
            break
    return new_size


def _adjust_batch_size(
    test_sample: Sequence[Union[torch.Tensor, Any]],
    batch_size: int,
    factor: float = 1.0,
    value: Optional[int] = None,
    **jac_kwargs,
) -> Tuple[int, bool]:
    max_batch_size = _get_max_batch_size(test_sample, **jac_kwargs)
    new_size = value if value is not None else int(batch_size * factor)
    new_size = min(new_size, max_batch_size)
    changed = new_size != batch_size
    return new_size, changed


def _get_max_batch_size(
    test_sample: Sequence[Union[torch.Tensor, Any]], **jac_kwargs
) -> int:
    argnums = jac_kwargs.get("argnums", 0)
    if isinstance(argnums, int):
        argnums = [argnums]
    batch_size = 0
    for argnum in argnums:
        arg = test_sample[argnum]
        assert isinstance(arg, torch.Tensor)
        batch_size += arg.numel()
    return batch_size



import warnings

import torch


try:
    import cupy.cuda
    from cupy_backends.cuda.libs import cublas, cusolver
except ImportError:
    cupy = None
    cusolver = None
    cublas = None


def psd_ridge(cov: torch.Tensor, rhs: torch.Tensor, penalty: float) -> torch.Tensor:
    """Solve ridge regression via Cholesky factorization, overwriting :attr:`cov` and :attr:`rhs`.

    Multiple right-hand sides are supported. Instead of providing the data
    matrix (commonly :math:`X` in ridge-regression notation), and labels (commonly :math:`y`),
    we are given directly :math:`\text{cov} = X^T X` and :math:`\text{rhs} = X^T y`.
    Since :attr:`cov` is symmetric only its **upper triangle** will be accessed.

    To limit memory usage, the :attr:`cov` matrix **may be overwritten**, and :math:`rhs`
    may also be overwritten (depending on its memory layout).

    Args:
        cov (Tensor): covariance of the linear system
        rhs (Tensor): right hand side (one or more) of the linear system
        penalty (float): Tikhonov l2 penalty

    Returns:
        solution (Tensor): the ridge regression coefficients
    """
    if cupy is not None and cov.device.type == "cuda":
        return _lowmem_psd_ridge(cov, rhs, penalty)
    else:
        # NOTE: this should be a warnings.warn NOT logger.warning - otherwise
        # it gets printed a lot of times and is just annoying. We could add
        # https://docs.python.org/library/logging.html#logging.captureWarnings
        # to the logger to capture warnings automatically.
        if cov.device.type == "cuda":
            warnings.warn(
                "low-memory solver cannot be used because `cupy` is not available. "
                "Install `cupy` if you encounter memory problems."
            )
        return _naive_psd_ridge(cov, rhs, penalty)


def _naive_psd_ridge(
    cov: torch.Tensor, rhs: torch.Tensor, penalty: float
) -> torch.Tensor:
    # Add diagonal without copies
    cov.diagonal().add_(penalty)
    # Solve with cholesky on GPU
    L = torch.linalg.cholesky(cov, upper=True)
    rhs_shape = rhs.shape
    return torch.cholesky_solve(rhs.view(cov.shape[0], -1), L, upper=True).view(
        rhs_shape
    )


def _lowmem_psd_ridge(
    cov: torch.Tensor, rhs: torch.Tensor, penalty: float
) -> torch.Tensor:
    assert cusolver is not None and cublas is not None and cupy is not None
    assert cov.device.type == "cuda"
    dtype = cov.dtype
    n = cov.shape[0]

    # Add diagonal without copies
    cov.diagonal().add_(penalty)

    if dtype == torch.float32:
        potrf = cusolver.spotrf
        potrf_bufferSize = cusolver.spotrf_bufferSize
        potrs = cusolver.spotrs
    elif dtype == torch.float64:
        potrf = cusolver.dpotrf
        potrf_bufferSize = cusolver.dpotrf_bufferSize
        potrs = cusolver.dpotrs
    else:
        raise ValueError(dtype)

    # cov must be f-contiguous (column-contiguous, stride is (1, n))
    assert cov.dim() == 2
    assert cov.shape[0] == cov.shape[1]
    transpose = False
    if n != 1:
        if cov.stride(0) != 1:
            cov = cov.T
            transpose = True
    assert cov.stride(0) == 1
    cov_cp = cupy.asarray(cov)

    # save rhs shape to restore it later on.
    rhs_shape = rhs.shape
    rhs = rhs.reshape(n, -1)
    n_rhs = rhs.shape[1]
    if rhs.stride(0) != 1:  # force rhs to be f-contiguous
        # `contiguous` causes a copy
        rhs = rhs.T.contiguous().T
    assert rhs.stride(0) == 1
    rhs_cp = cupy.asarray(rhs)

    handle = cupy.cuda.device.get_cusolver_handle()
    uplo = cublas.CUBLAS_FILL_MODE_LOWER if transpose else cublas.CUBLAS_FILL_MODE_UPPER
    dev_info = torch.empty(
        1, dtype=torch.int32
    )  # don't allocate with cupy as it uses a separate mem pool
    dev_info_cp = cupy.asarray(dev_info)

    worksize = potrf_bufferSize(handle, uplo, n, cov_cp.data.ptr, n)
    workspace = torch.empty(worksize, dtype=dtype)
    workspace_cp = cupy.asarray(workspace)

    # Cholesky factorization
    potrf(
        handle,
        uplo,
        n,
        cov_cp.data.ptr,
        n,
        workspace_cp.data.ptr,
        worksize,
        dev_info_cp.data.ptr,
    )
    if (dev_info_cp != 0).any():
        raise torch.linalg.LinAlgError(
            f"Error reported by {potrf.__name__} in cuSOLVER. devInfo = {dev_info_cp}."
        )

    # Solve: A * X = B
    potrs(
        handle,
        uplo,
        n,
        n_rhs,
        cov_cp.data.ptr,
        n,
        rhs_cp.data.ptr,
        n,
        dev_info_cp.data.ptr,
    )
    if (dev_info_cp != 0).any():
        raise torch.linalg.LinAlgError(
            f"Error reported by {potrf.__name__} in cuSOLVER. devInfo = {dev_info_cp}."
        )

    return torch.as_tensor(rhs).reshape(rhs_shape)



from typing import Mapping
import torch

import franken.utils.distributed as dist_utils
from franken.data.base import Target


class BaseMetric:
    def __init__(
        self,
        name: str,
        device: torch.device,
        dtype: torch.dtype = torch.float64,
        units: Mapping[str, str | None] = {},
    ):
        self.name = name
        self.device = device
        self.dtype = dtype
        self.buffer = None
        self.samples_counter = torch.zeros((1,), device=device, dtype=dtype)
        self.units = units

    def reset(self) -> None:
        """Reset the buffer to zeros"""
        self.buffer = None
        self.samples_counter = torch.zeros((1,), device=self.device, dtype=torch.int64)

    def buffer_add(self, value: torch.Tensor, num_samples: int = 1) -> None:
        if self.buffer is None:
            self.buffer = torch.zeros(value.shape, device=self.device, dtype=self.dtype)
        else:
            assert self.buffer.shape == value.shape
        self.buffer += value
        self.samples_counter += num_samples

    def update(
        self,
        predictions: Target,
        targets: Target,
    ) -> None:
        """Update the metric buffer with new batch results"""
        raise NotImplementedError()

    def compute(self, reset: bool = True) -> torch.Tensor:
        if self.buffer is None:
            raise ValueError(
                f"Cannot compute value for metric '{self.name}' "
                "because it was never updated."
            )
        dist_utils.all_sum(self.buffer)
        dist_utils.all_sum(self.samples_counter)
        error = self.buffer / self.samples_counter
        if reset:
            self.reset()
        return error



from math import ceil, sqrt

import numpy as np
import scipy as sp
import torch
import torch.utils.data


class RandomFeaturesHead(torch.nn.Module):
    r"""Base class for Random Feature heads

    Args:
        input_dim (int): Dimension of the input features.
        num_random_features (int, optional): The number of random features to use in the feature mapping. Defaults to :math:`2^{10} = 1024`.
        num_species (int | None): The number of chemical species for
            which the kernel is computed. This parameter is relevant for systems
            with multiple chemical species. Defaults to :code:`None`.
        chemically_informed_ratio (float | None): The relative weight of chemically-informed kernels with respect to the all-species kernel. Ignored if
            :code:`num_species` is None. Defaults to :code:`None`.
    """

    def __init__(
        self,
        input_dim: int,
        num_random_features: int = 2**10,
        num_species: int | None = None,
        chemically_informed_ratio: float | None = None,
    ):
        super(RandomFeaturesHead, self).__init__()
        self.input_dim = input_dim
        self.num_random_features = num_random_features
        self.num_species = num_species
        self.chemically_informed_ratio = chemically_informed_ratio

        self._per_species_multiplier = 0
        if self.num_species is not None:
            assert self.num_species > 0
            self._per_species_multiplier += self.num_species
            if self.chemically_informed_ratio is not None:
                self._per_species_multiplier += 1
        else:
            self._per_species_multiplier += 1
        self.total_random_features = (
            self._per_species_multiplier * self.num_random_features
        )

        # Register weights buffer
        self.weights: torch.Tensor
        self.register_buffer("weights", torch.zeros((1, self.total_random_features)))

    def species_scatter_sum(
        self, Z: torch.Tensor, atomic_numbers: torch.Tensor | None = None
    ) -> torch.Tensor:
        if self.num_species is None:
            return Z.mean(0)
        else:
            assert (
                atomic_numbers is not None
            ), "atomic_number should be specified when self.num_species is not None"
            species_map, scatter_idxs = torch.unique(
                atomic_numbers, sorted=True, return_inverse=True
            )
            assert (
                len(species_map) == self.num_species
            ), f"The provided atomic numbers {species_map}, are of a different number than self.num_species {self.num_species}"

            scatter_idxs = scatter_idxs.unsqueeze(-1).expand(
                Z.size()
            )  # ~[natoms, random_features_per_species] Broadcasting for backprop.
            chemically_informed_descriptors = (
                torch.zeros((self.num_species, self.num_random_features), dtype=Z.dtype)
                .to(Z.device)
                .scatter_reduce_(0, scatter_idxs, Z, "mean")
            ) / self.num_species  # Normalize.

            if self.chemically_informed_ratio is not None:
                kappa = self.chemically_informed_ratio
                assert (
                    0 <= kappa <= 1
                ), "The ratio of chemically informed feature feature map should be bounded between 0 and 1"
                chemically_informed_descriptors = torch.cat(
                    (
                        (1 - kappa) * Z.mean(0, keepdim=True),
                        kappa * chemically_informed_descriptors,
                    )
                )
            return chemically_informed_descriptors.view(-1)

    def init_args(self):
        return {
            "num_random_features": self.num_random_features,
            "num_species": self.num_species,
            "chemically_informed_ratio": self.chemically_informed_ratio,
        }


class OrthogonalRFF(RandomFeaturesHead):
    r"""
    `Orthogonal Random Fourier Features <https://arxiv.org/abs/1610.09072>`_ by Yu et al. for approximating the gaussian kernel

    .. math::
        \text{exp}\left(-\frac{\| x - y \|^{2}}{2\ell^{2}}\right)


    Args:
        input_dim (int): Dimensionality of the input features.
        num_random_features (int, optional): The number of random features to use in the feature mapping. Defaults to :math:`2^{10} = 1024`.
        num_species (int | None): The number of chemical species for
            which the kernel is computed. This parameter is relevant for systems
            with multiple chemical species. Defaults to :code:`None`.
        chemically_informed_ratio (float | None): The relative weight of chemically-informed kernels with respect to the all-species kernel. Ignored if
            :code:`num_species` is None. Defaults to :code:`None`.
        use_offset (bool): A flag indicating whether to use an offset in
            the random feature generation. Using an offset reduces the number of
            random features by half but increases variance. Defaults to :code:`True`.
        length_scale (float): The length scale parameter :math:`\ell` that
            controls the smoothness of the kernel function. It affects how quickly
            the kernel values decay with distance. Defaults to 1.0.
        rng_seed (int | None): A seed for the random number generator
            used in generating random features. Setting this ensures reproducibility
            of results. Defaults to :code:`None`.
    """

    def __init__(
        self,
        input_dim: int,
        num_random_features: int = 2**10,
        num_species: int | None = None,
        chemically_informed_ratio: float | None = None,
        use_offset: bool = True,
        length_scale: float = 1.0,
        rng_seed=None,
    ):
        check_positive_arg(length_scale, "length_scale")
        super(OrthogonalRFF, self).__init__(
            input_dim,
            num_random_features,
            num_species=num_species,
            chemically_informed_ratio=chemically_informed_ratio,
        )
        self.rng_seed = rng_seed
        random_generator = get_seeded_generator(self.rng_seed)

        W = self._sample_orf_matrix(
            input_dim, self.num_random_features, random_generator, self.rng_seed
        )

        self.register_buffer("rff_matrix", W)
        self.register_buffer(
            "random_offset",
            torch.rand(self.num_random_features, generator=random_generator)
            * 2
            * np.pi,
        )
        self.register_buffer("length_scale", torch.tensor(length_scale))

        self.use_offset = use_offset
        if not self.use_offset:
            self.total_random_features = self.total_random_features * 2

    @staticmethod
    def _sample_orf_matrix(input_dim, num_random_features, torch_generator, rng_seed):
        num_folds = int(np.ceil(num_random_features / input_dim))

        G = torch.randn(
            num_folds,
            input_dim,
            input_dim,
            generator=torch_generator,
        )
        Q, _ = torch.linalg.qr(
            G, mode="complete"
        )  # The _columns_ in each batch of matrices in Q are orthonormal.

        Q = Q.transpose(
            -1, -2
        )  # The _rows_ in each batch of matrices in Q are orthonormal.

        S = torch.tensor(
            sp.stats.chi.rvs(
                input_dim,
                size=(num_folds, input_dim, 1),
                random_state=rng_seed,
            ),
        )

        W = Q * S  # [num_folds, input_dim, input_dim]
        W = torch.cat(
            [W[fold_idx, ...] for fold_idx in range(num_folds)], dim=0
        )  # Concatenate columns [num_heads, num_folds*input_dim, input_dim]
        W = W[:num_random_features, :]
        return W

    def feature_map(
        self,
        h: torch.Tensor,
        atomic_numbers: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """
        h: torch.Tensor - descriptors for a single configuration ~[natoms, descriptors]
        atomic_numbers: torch.Tensor - atomic numbers for a single configuration ~[natoms]
        """
        length_scale = self.length_scale
        # Convert dtype of rf matrices to same as h
        self.rff_matrix = self.rff_matrix.to(dtype=h.dtype)
        if self.use_offset:
            self.random_offset = self.random_offset.to(dtype=h.dtype)
        # Compute the orthogonal random features nonlinear mapping
        Z = torch.einsum("af, Ff -> aF", h, self.rff_matrix / length_scale)
        if self.use_offset:
            b = self.random_offset.unsqueeze(0)
            Z = torch.cos(Z + b) / sqrt(0.5 * self.num_random_features)
        else:
            _s = torch.sin(Z)
            _c = torch.cos(Z)
            Z = torch.cat((_s, _c), dim=-1) / sqrt(self.num_random_features)

        return self.species_scatter_sum(
            Z, atomic_numbers=atomic_numbers
        )  # Sum over the single species

    def init_args(self):
        return super().init_args() | {
            "use_offset": self.use_offset,
            "length_scale": self.length_scale.tolist(),
            "rng_seed": self.rng_seed,
        }


class BiasedOrthogonalRFF(OrthogonalRFF):
    def __init__(
        self,
        input_dim: int,
        num_random_features: int = 2**10,
        num_species: int | None = None,
        chemically_informed_ratio: float | None = None,
        use_offset: bool = True,
        length_scale: float = 1.0,
        bias: float = 1.0,
        rng_seed=None,
    ):
        super(BiasedOrthogonalRFF, self).__init__(
            input_dim,
            num_random_features,
            num_species=num_species,
            chemically_informed_ratio=chemically_informed_ratio,
            use_offset=use_offset,
            length_scale=length_scale,
            rng_seed=rng_seed,
        )
        self.register_buffer("bias", torch.tensor(bias))

    def feature_map(
        self,
        h: torch.Tensor,
        atomic_numbers: torch.Tensor | None = None,
    ) -> torch.Tensor:
        self.bias = self.bias.to(dtype=h.dtype)
        # unbiased_features ~ [num_random_features - 1]
        features = super().feature_map(h, atomic_numbers=atomic_numbers)
        assert features.ndim == 1
        features[-1] = torch.sqrt(self.bias)
        return features

    def init_args(self):
        return super().init_args() | {"bias": self.bias.tolist()}


class MultiScaleOrthogonalRFF(RandomFeaturesHead):
    def __init__(
        self,
        input_dim: int,
        num_random_features: int = 2**10,
        num_species: int | None = None,
        chemically_informed_ratio: float | None = None,
        use_offset: bool = True,
        length_scale_low: float = 1.0,
        length_scale_high: float = 10.0,
        length_scale_num: int = 4,
        rng_seed=None,
    ):
        super(MultiScaleOrthogonalRFF, self).__init__(
            input_dim,
            num_random_features,
            num_species=num_species,
            chemically_informed_ratio=chemically_informed_ratio,
        )
        self.rng_seed = rng_seed
        random_generator = get_seeded_generator(self.rng_seed)
        check_positive_arg(length_scale_low, "length_scale_low")
        self.length_scale_low = length_scale_low
        check_positive_arg(length_scale_high, "length_scale_high")
        self.length_scale_high = length_scale_high
        self.length_scale_num = length_scale_num

        W = OrthogonalRFF._sample_orf_matrix(
            input_dim, self.num_random_features, random_generator, self.rng_seed
        )

        self.register_buffer("rff_matrix", W)
        self.register_buffer(
            "random_offset",
            torch.rand(self.num_random_features, generator=random_generator)
            * 2
            * np.pi,
        )

        _num_repeats = ceil(self.num_random_features / self.length_scale_num)
        _length_scale = torch.repeat_interleave(
            torch.linspace(
                self.length_scale_low, self.length_scale_high, self.length_scale_num
            ),
            _num_repeats,
        )
        self.register_buffer("length_scale", _length_scale[: self.num_random_features])

        self.use_offset = use_offset
        if not self.use_offset:
            self.total_random_features = self.total_random_features * 2

    def feature_map(
        self,
        h: torch.Tensor,
        atomic_numbers: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """
        h: torch.Tensor - descriptors for a single configuration ~[natoms, descriptors]
        atomic_numbers: torch.Tensor - atomic numbers for a single configuration ~[natoms]
        """
        # Convert dtype of rf matrices to same as h
        self.rff_matrix = self.rff_matrix.to(dtype=h.dtype)
        if self.use_offset:
            self.random_offset = self.random_offset.to(dtype=h.dtype)

        self.length_scale = self.length_scale.to(dtype=h.dtype)
        length_scale = self.length_scale.view(-1, 1)
        # Compute the orthogonal random features nonlinear mapping
        Z = torch.einsum("af, Ff -> aF", h, self.rff_matrix / length_scale)
        if self.use_offset:
            b = self.random_offset.unsqueeze(0)
            Z = torch.cos(Z + b) / sqrt(0.5 * self.num_random_features)
        else:
            _s = torch.sin(Z)
            _c = torch.cos(Z)
            Z = torch.cat((_s, _c), dim=-1) / sqrt(self.num_random_features)

        return self.species_scatter_sum(
            Z, atomic_numbers=atomic_numbers
        )  # Sum over the single species

    def init_args(self):
        return super().init_args() | {
            "use_offset": self.use_offset,
            "length_scale_low": self.length_scale_low,
            "length_scale_high": self.length_scale_high,
            "length_scale_num": self.length_scale_num,
            "rng_seed": self.rng_seed,
        }


class Linear(RandomFeaturesHead):
    r"""
    Module implementing the linear kernel

    .. math::
        \frac{\langle x, y \rangle}{\ell^{2}} + b


    Args:
        input_dim (int): The dimensionality of the input features.
        num_species (int | None): The number of chemical species for
            which the kernel is computed. This parameter is relevant for systems
            with multiple chemical species. Defaults to :code:`None`.
        chemically_informed_ratio (float | None): The relative weight of chemically-informed kernels with respect to the all-species kernel. Ignored if
            :code:`num_species` is None. Defaults to :code:`None`.
        bias (float): The bias term :math:`b` added to the kernel function. This allows for shifting the kernel values, which can be useful for certain applications. Defaults to 0.0.
        length_scale (float): The length scale parameter :math:`\ell` that controls the smoothness of the kernel function. It affects how quickly the kernel values decay with distance. Defaults to 1.0.
    """

    def __init__(
        self,
        input_dim: int,
        num_species: int | None = None,
        chemically_informed_ratio: float | None = None,
        bias: float = 0.0,
        length_scale: float = 1.0,
    ):
        check_positive_arg(length_scale, "length_scale")
        super(Linear, self).__init__(
            input_dim,
            # Add a constant bias feature
            input_dim + 1,
            num_species=num_species,
            chemically_informed_ratio=chemically_informed_ratio,
        )

        self.register_buffer("bias", torch.tensor(bias))
        self.register_buffer("length_scale", torch.tensor(length_scale))

    def feature_map(
        self,
        h: torch.Tensor,
        atomic_numbers: torch.Tensor | None = None,
    ) -> torch.Tensor:
        length_scale = self.length_scale
        scaled_descriptors = h / length_scale
        # Add a constant bias feature to scaled_descriptors
        bias_feature = (
            torch.sqrt(self.bias)
            .expand(scaled_descriptors.size(0), 1)
            .to(dtype=h.dtype)
        )
        scaled_descriptors = torch.cat([scaled_descriptors, bias_feature], dim=1)

        return self.species_scatter_sum(
            scaled_descriptors, atomic_numbers=atomic_numbers
        )

    def init_args(self):
        _super_init = super().init_args()
        _super_init.pop("num_random_features")
        return _super_init | {
            "bias": self.bias.tolist(),
            "length_scale": self.length_scale.tolist(),
        }


class TensorSketch(RandomFeaturesHead):
    r"""
    TensorSketch of the polynomial kernel

    .. math::
        \left(\frac{\langle x, y \rangle}{\ell^{2}} + b\right)^{d}


    from `"Fast and scalable polynomial kernels via explicit feature maps" <https://doi.org/10.1145/2487575.2487591>`_ by Pham and Pagh.

    Args:
        input_dim (int): The dimensionality of the input features.
        num_random_features (int): The number of random features to use in the sketching process. Defaults to :math:`2^{10} = 1024`.
        num_species (int | None): The number of chemical species for
            which the kernel is computed. This parameter is relevant for systems
            with multiple chemical species. Defaults to :code:`None`.
        chemically_informed_ratio (float | None): The relative weight of chemically-informed kernels with respect to the all-species kernel. Ignored if
            :code:`num_species` is None. Defaults to :code:`None`.
        degree (int): Degree :math:`d` of the polynomial kernel to sketch. Defaults to 2.
        bias (float): The bias term :math:`b` added to the kernel function. This allows for shifting the kernel values, which can be useful for certain applications. Defaults to 0.0.
        length_scale (float): The length scale parameter :math:`\ell` that controls the smoothness of the kernel function. It affects how quickly the kernel values decay with distance. Defaults to 1.0.
        rng_seed (int | None): A seed for the random number generator used in generating random features. Setting this ensures reproducibility of results. Defaults to :code:`None`.
    """

    def __init__(
        self,
        input_dim: int,
        num_random_features: int = 2**10,
        num_species: int | None = None,
        chemically_informed_ratio: float | None = None,
        degree=2,
        bias: float = 0.0,
        length_scale: float = 1.0,
        rng_seed=None,
    ):
        check_positive_arg(length_scale, "length_scale")
        super(TensorSketch, self).__init__(
            input_dim,
            num_random_features,
            num_species=num_species,
            chemically_informed_ratio=chemically_informed_ratio,
        )
        self.rng_seed = rng_seed
        self.register_buffer("degree", torch.tensor(degree))
        self.register_buffer("bias", torch.tensor(bias))
        self.register_buffer("length_scale", torch.tensor(length_scale))
        self.register_buffer(
            "count_sketches", self._sample_count_sketches(self.rng_seed, self.degree)
        )

    def _sample_count_sketches(self, rng_seed, degree: int):
        rng_torch = get_seeded_generator(rng_seed)
        with torch.no_grad():
            count_sketches = []
            for _ in range(degree):
                count_sketches.append(self._sample_hash_fns(rng_torch))
            return torch.stack(count_sketches)  # ~[degree, random_features, input_dim]

    def _sample_hash_fns(self, generator: torch.Generator | None):
        i_hash = torch.randint(
            low=0,
            high=self.num_random_features,
            size=(self.input_dim,),
            generator=generator,
        )
        s_hash = torch.bernoulli(
            0.5 * torch.ones(self.input_dim), generator=generator
        )  # Bernoulli
        # Convert to Rademacher
        s_hash = s_hash * 2 - 1
        indices = torch.stack([i_hash, torch.arange(self.input_dim)], dim=0)
        count_sketch = torch.sparse_coo_tensor(
            indices,
            s_hash,
            torch.Size([self.num_random_features, self.input_dim]),
        ).to_dense()
        return count_sketch

    def feature_map(
        self,
        h: torch.Tensor,
        atomic_numbers: torch.Tensor | None = None,
    ) -> torch.Tensor:
        length_scale = self.length_scale
        scaled_descriptors = (h + torch.sqrt(torch.abs(self.bias))) / length_scale
        # Convert dtype of self to the same dtype as h.
        # Note: cannot use self.to(dtype) due to self.degree buffer being int!
        self.count_sketches = self.count_sketches.to(dtype=h.dtype)
        Z = torch.einsum(
            "af, dFf -> adF", scaled_descriptors, self.count_sketches
        )  # ~[atoms, degree, random_features]
        Z = torch.fft.rfft(Z, n=self.num_random_features)
        Z = Z.prod(1)  # ~[atoms, random_features]
        Z = torch.fft.irfft(Z, n=self.num_random_features)
        return self.species_scatter_sum(Z, atomic_numbers=atomic_numbers)

    def init_args(self):
        return super().init_args() | {
            "degree": self.degree.tolist(),
            "bias": self.bias.tolist(),
            "length_scale": self.length_scale.tolist(),
            "rng_seed": self.rng_seed,
        }


def check_positive_arg(value, name):
    if value <= 0:
        raise ValueError(f"Argument {name} must be positive. Got {value}.")


def get_seeded_generator(seed: int | None) -> torch.Generator | None:
    rng_torch = None
    if seed is not None:
        rng_torch = torch.Generator(device=torch.get_default_device())
        rng_torch.manual_seed(seed)
    return rng_torch



from typing import Literal, Sequence
import itertools
import json
import logging

from ase import Atoms
from ase.geometry.rdf import get_recommended_r_max
from ase.io import read
import ase.data
import hydra
import numpy as np
from pathlib import Path
from omegaconf import DictConfig

from franken.backbones.utils import CacheDir
from franken.mdgen.diffusion_coeff import calc_self_diffusion, calc_water_msd
from franken.mdgen.mdgen_utils import register_hydra_resolvers
from franken.utils.misc import pprint_config, setup_logger
from franken.mdgen.rdf_utils import (
    distance_matrix,
    avg_rdf,
    compute_rdf_mae,
    stability_rdf,
)

logger = logging.getLogger("franken")


class NumpyEncoder(json.JSONEncoder):
    """JSON encoder to dump numpy arrays as python lists"""

    def default(self, o):
        if isinstance(o, np.ndarray):
            return o.tolist()
        return super().default(o)


def read_traj(path) -> list[Atoms]:
    traj = read(path, index=":")
    if isinstance(traj, Atoms):
        traj = [traj]
    return traj


def get_rmax(*trajs: list[Atoms]) -> float:
    rmax = float("inf")
    for traj in trajs:
        for frame in traj:
            rmax = min(rmax, get_recommended_r_max(frame.cell, frame.pbc))  # type: ignore
    return rmax


def postprocessing_water_diffusion_coef(
    traj_paths: list[Path], timestep_fs: float, t_min_fs: float = 0.0
) -> dict[str, dict]:
    # Loop over trajectories
    results = {}
    for traj_path in traj_paths:
        name = traj_path.stem
        if not traj_path.exists():
            logger.warning(f"Trajectory {name} does not exist. Skipping.")
            continue
        logger.info(f"Water diffusion coefficient for trajectory {name}...")
        results[name] = {}

        traj = read_traj(traj_path)

        msd = calc_water_msd(traj)
        ds, linmods = calc_self_diffusion(
            msd, delta_t_fs=timestep_fs, t_min=None, t_min_fs=t_min_fs
        )
        # Report all possibly useful data in a serializable way
        results[name]["MSD"] = msd
        results[name]["D"] = ds
        results[name]["slope"] = [lm.slope for lm in linmods]  # type: ignore
        results[name]["intercept"] = [lm.intercept for lm in linmods]  # type: ignore
        results[name]["stderr"] = [lm.stderr for lm in linmods]  # type: ignore
        results[name]["intercept_stderr"] = [lm.intercept_stderr for lm in linmods]  # type: ignore
    return results


def postprocessing_rdf(
    traj_paths: list[Path],
    ref_traj_path: Path | None,
    timestep_fs: float,
    nbins_per_angstrom=50,
    stability_threshold=1.0,
    stability_interval: int = 20,
    equilibration_fs: float = 200,
    rdf_version: Literal["fast", "ase"] = "fast",
    distance_version: Literal["fast", "ase"] = "fast",
) -> dict[str, dict]:
    # Load reference
    ref_traj, ref_distance_matrix = None, None
    if ref_traj_path is not None:
        ref_traj = read_traj(ref_traj_path)
        ref_distance_matrix = distance_matrix(ref_traj, version=distance_version)

    # Loop over trajectories
    results = {}
    for traj_path in traj_paths:
        name = traj_path.stem
        if not traj_path.exists():
            logger.warning(f"Trajectory {name} does not exist. Skipping.")
            continue
        logger.info(f"RDF analysis for trajectory {name}...")
        results[name] = {}

        # Load traj and compute distances
        traj = read_traj(traj_path)
        # Skip first time-step which is just the initial configuration
        traj = traj[1:]
        # Skip the beginning of the trajectory where simulation is not at equilibrium
        traj_len_fs = timestep_fs * len(traj)
        if traj_len_fs <= equilibration_fs:
            # Did not reach equilibrium time
            results[name]["total_traj_time"] = 0
            results[name]["total_traj_time_fs"] = 0
            continue
        else:
            traj = traj[int(equilibration_fs / timestep_fs) :]
        results[name]["total_traj_time"] = len(traj)
        results[name]["total_traj_time_fs"] = len(traj) * timestep_fs
        traj_distance_matrix = distance_matrix(traj, version=distance_version)

        # Rmax and nbins for this trajectory
        if ref_traj is not None:
            rmax = get_rmax(ref_traj, traj)
        else:
            rmax = get_rmax(traj)
        nbins = int(rmax * nbins_per_angstrom)

        unique_atomic_numbers = np.unique(traj[0].numbers)
        # Since the inter-atomic distance is symmetric we don't care about ordering i.e. (1, 8), (8, 1)
        # therefore don't use carthesian product, get all length 2 combinations using this other
        # itertools function
        element_pairs = list(
            itertools.combinations_with_replacement(unique_atomic_numbers, 2)
        )
        for el_pair in element_pairs:
            el_pair = tuple(sorted(el_pair))  # for consistency
            el_pair_str = f"{ase.data.chemical_symbols[el_pair[0]]}-{ase.data.chemical_symbols[el_pair[1]]}"
            # Full RDF for test trajectory
            full_traj_rdf, radii = avg_rdf(
                traj,
                distance_matrix=traj_distance_matrix,
                rmax=rmax,
                nbins=nbins,
                elements=el_pair,
                version=rdf_version,
            )
            results[name][f"{el_pair_str}_rdf"] = full_traj_rdf
            # radii won't change with different elements.
            results[name]["dists"] = radii
            # Stability of the trajectory with respect to the average of the trajectory itself
            stability_time_self, stability_mae_self = stability_rdf(
                traj=traj,
                dist=traj_distance_matrix,
                ref_rdf=full_traj_rdf,
                rmax=rmax,
                nbins=nbins,
                stability_interval=stability_interval,
                threshold=stability_threshold,
                elements=el_pair,
                rdf_version=rdf_version,
            )
            results[name][f"{el_pair_str}_stability_time_self"] = stability_time_self
            results[name][f"{el_pair_str}_stability_mae_self"] = stability_mae_self

            # Stability of the trajectory with respect to the reference trajectory
            if ref_traj is not None:
                assert ref_distance_matrix is not None
                # Compute reference RDF. Recompute needed since rmax may have changed
                ref_rdf = avg_rdf(
                    ref_traj,
                    distance_matrix=ref_distance_matrix,
                    rmax=rmax,
                    nbins=nbins,
                    no_dists=True,
                    elements=el_pair,
                    version=rdf_version,
                )
                assert isinstance(ref_rdf, np.ndarray)
                results[name][f"{el_pair_str}_ref_rdf"] = ref_rdf
                full_mae = compute_rdf_mae(
                    full_traj_rdf, ref_rdf, delta_r=radii[1] - radii[0]
                )
                results[name][f"{el_pair_str}_rdf_mae"] = full_mae
                stability_time_ref, stability_mae_ref = stability_rdf(
                    traj=traj,
                    dist=traj_distance_matrix,
                    ref_rdf=ref_rdf,
                    rmax=rmax,
                    nbins=nbins,
                    stability_interval=stability_interval,
                    threshold=stability_threshold,
                    elements=el_pair,
                    rdf_version=rdf_version,
                )
                results[name][f"{el_pair_str}_stability_time_ref"] = stability_time_ref
                results[name][f"{el_pair_str}_stability_mae_ref"] = stability_mae_ref
    return results


def save_analysis_results(
    traj_paths: list[Path], results: Sequence[dict], save_full: bool = True
) -> None:
    # Check that the results are consistent (all analyses have same 'trajectory-name' keys)
    valid_names = set(results[0].keys())
    for anl_res in results:
        if set(anl_res.keys()) != valid_names:
            logger.error(
                f"Cannot save postprocessing results. Trajectory names differ between analyses. "
                f"Valid file-names are {valid_names}, found names {set(anl_res.keys())}."
            )
            return

    # Save the analyses for individual trajectories
    save_folder = None
    merged_analysis = {}
    for traj_path in traj_paths:
        name = traj_path.stem

        save_folder = traj_path.parent
        save_path = save_folder / f"{name}_analysis.json"
        with open(save_path, "w") as f:
            merged_traj_analysis = {
                k: v for anl_res in results for k, v in anl_res[name].items()
            }
            merged_analysis[name] = merged_traj_analysis
            json.dump(merged_traj_analysis, f, cls=NumpyEncoder)
        logger.info(f"Saved postprocessing results to: '{save_path.resolve()}'")

    # save the whole analysis as well.
    if save_full and save_folder is not None:
        if not all(traj_paths[0].parent == tp.parent for tp in traj_paths):
            logger.warning(
                "Cannot save full analysis because trajectories are not in the same folder."
            )
        else:
            save_path = traj_paths[0].parent / "analysis.json"
            if save_path.exists():
                logger.warning(
                    f"Full analysis save-path '{save_path.resolve()}' exists and will be overwritten."
                )
            with open(save_path, "w") as f:
                json.dump(merged_analysis, f, cls=NumpyEncoder)
            logger.info(
                f"Saved whole postprocessing results to: '{save_path.resolve()}'"
            )


@hydra.main(
    version_base="1.3", config_path="configs", config_name="md_postprocess.yaml"
)
def md_postprocess(cfg: DictConfig):
    output_dir = Path(cfg.paths.run_dir)
    logging_level = cfg.console_logging_level.upper()
    setup_logger(level=logging_level, directory=output_dir, logname="asemd")
    pprint_config(cfg)
    if not CacheDir.is_initialized():
        CacheDir.initialize()

    # Find traj paths
    traj_ext = "extxyz"
    trajectories = list(output_dir.glob(f"*.{traj_ext}"))
    if len(trajectories) == 0:
        raise RuntimeError(
            f"No trajectory with extension {traj_ext} found "
            f"in output directory {output_dir}"
        )
    logger.info(f"Found {len(trajectories)} to postprocess")

    all_analyses = []
    post_configs = cfg.metrics
    for post_name, post_cfg in post_configs.items():
        logger.info(f"Running {post_name} postprocessing...")
        all_analyses.append(hydra.utils.instantiate(post_cfg, traj_paths=trajectories))
    if len(all_analyses) > 0:
        save_analysis_results(traj_paths=trajectories, results=all_analyses)

    logger.info("Postprocessing complete.")


def pp_cli_entry_point():
    register_hydra_resolvers()
    md_postprocess()


if __name__ == "__main__":
    pp_cli_entry_point()



from pathlib import Path

import torch
from fairchem.core.preprocessing import AtomsToGraphs as FairchemAtomsToGraphs

import franken.utils.distributed as dist_utils
from franken.backbones import REGISTRY
from franken.backbones.utils import get_checkpoint_path
from franken.data import BaseAtomsDataset, Configuration, Target


class FairchemAtomsDataset(BaseAtomsDataset):
    def __init__(
        self,
        data_path: str | Path | None,
        split: str,
        num_random_subsamples: int | None = None,
        subsample_rng: int | None = None,
        gnn_backbone_id: str | None = None,
        cutoff=6.0,
        max_num_neighbors=200,
        precompute=True,
    ):
        super().__init__(data_path, split, num_random_subsamples, subsample_rng)

        if gnn_backbone_id is not None:
            cutoff, max_num_neighbors = self.load_info_from_gnn_config(gnn_backbone_id)

        if split == "md":
            # Cannot get energy and forces in MD mode (the calculator fails)
            self.a2g = FairchemAtomsToGraphs(
                max_neigh=max_num_neighbors,
                radius=cutoff,  # type: ignore
            )
        else:
            self.a2g = FairchemAtomsToGraphs(
                max_neigh=max_num_neighbors,
                radius=cutoff,
                r_energy=True,
                r_forces=True,  # type: ignore
            )
        self.graphs = None
        if precompute and len(self.ase_atoms) > 0:
            self.graphs = self.a2g.convert_all(
                self.ase_atoms, disable_tqdm=dist_utils.get_rank() != 0
            )

    def load_info_from_gnn_config(self, gnn_backbone_id: str):
        ckpt_path = get_checkpoint_path(gnn_backbone_id)
        backbone_family = REGISTRY[gnn_backbone_id]["kind"].lower()
        if backbone_family != "fairchem":
            raise ValueError(f"Incompatible backbone family {backbone_family}")
        model_cfg = torch.load(ckpt_path, map_location="cpu", weights_only=False)[
            "config"
        ]["model"]
        cutoff = getattr(model_cfg, "cutoff", 6.0)
        max_num_neighbors = getattr(model_cfg, "max_num_neighbors", 200)
        del model_cfg
        return cutoff, max_num_neighbors

    def graph_to_inputs(self, graph):
        return Configuration(
            atom_pos=graph.pos,  # type: ignore
            atomic_numbers=graph.atomic_numbers.int(),
            natoms=torch.tensor(graph.natoms).view(1),
            batch_ids=(
                graph.batch
                if graph.batch is not None
                else torch.zeros(graph.natoms, dtype=torch.int64)
            ),
            cell=graph.cell,
            pbc=getattr(graph, "pbc", None),
        )

    def graph_to_targets(self, graph):
        energy = torch.tensor(graph.energy)
        return Target(energy=energy, forces=graph.forces)

    def __getitem__(self, idx, no_targets: bool = False):
        """Returns an array of (inputs, outputs) with inputs being a configuration
        and outputs being the target (energy and forces).
        Note: ONLY for the 'train' split, the energy_shift is removed from the target.
        """
        if self.graphs is None:
            graph = self.a2g.convert(self.ase_atoms[idx])
        else:
            graph = self.graphs[idx]

        data = self.graph_to_inputs(graph)
        if no_targets:
            return data
        target = self.graph_to_targets(graph)

        if self.split == "train":
            target.energy -= self.energy_shifts[idx]

        return data, target



import datetime
import json
import logging
import warnings
from pathlib import Path
from typing import Any, Dict, Literal, NamedTuple, Optional, Tuple
from uuid import uuid4

import hydra
import torch.distributed
import torch.utils.data
from omegaconf import DictConfig, OmegaConf

import franken.utils.distributed as dist_utils
from franken import FRANKEN_DIR
from franken.backbones.utils import CacheDir, download_checkpoint
from franken.data import BaseAtomsDataset
from franken.rf.model import FrankenPotential
from franken.trainers import BaseTrainer
from franken.trainers.log_utils import DataSplit, LogEntry
from franken.utils.misc import (
    garbage_collection_cuda,
    get_device_name,
    params_grid,
    pprint_config,
    setup_logger,
)

DATASETS_PATH = FRANKEN_DIR / "datasets"
with open(DATASETS_PATH / "registry.json", "r") as f:
    DATASET_REGISTRY = json.load(f)


class BestTrial(NamedTuple):
    trial_id: int
    log: LogEntry


warnings.filterwarnings(
    "ignore",
    message=r"You are using `torch.load` with `weights_only=False`",
    category=FutureWarning,
)
warnings.filterwarnings(
    "ignore",
    message=r"`torch.cuda.amp.autocast\(args...\)` is deprecated",
    category=FutureWarning,
)
logger = logging.getLogger("franken")


def init_loaders(
    gnn_backbone_id: str,
    train_path: Optional[Path],
    val_path: Optional[Path] = None,
    test_path: Optional[Path] = None,
    num_train_subsamples: int | None = None,
    subsample_rng: int | None = None,
) -> Dict[str, torch.utils.data.DataLoader]:
    datasets: Dict[str, BaseAtomsDataset] = {}
    for split, data_path in zip(
        ["train", "val", "test"], [train_path, val_path, test_path]
    ):
        if data_path is not None:
            dset = BaseAtomsDataset.from_path(
                data_path=data_path,
                split=split,
                gnn_backbone_id=gnn_backbone_id,
                num_random_subsamples=(
                    num_train_subsamples if split == "train" else None
                ),
                subsample_rng=subsample_rng,
            )
            datasets[split] = dset

    dataloaders = {
        split: dset.get_dataloader(distributed=torch.distributed.is_initialized())
        for split, dset in datasets.items()
    }
    return dataloaders


def create_params_grid(
    group_name: Literal["random_features", "solver"],
    group: DictConfig,
):
    group_dict = {}

    for k, v in group.items():
        if k == "num_random_features":
            assert isinstance(
                v, int
            ), f"The number of random features should not be optimized over. Got num_random_features={v} {type(v)}"

        if isinstance(v, DictConfig) and "_target_" in v:
            v = hydra.utils.instantiate(v)
        elif isinstance(v, (int, float)):
            v = [v]

        group_dict[k] = v

    if group_name == "solver":
        return group_dict
    else:
        return params_grid(group_dict)


def filter_rf_hp_grid(rf_hp: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    # if num_species is None, chemically_informed_ratio must also be None
    if rf_hp.get("num_species", None) is None:
        if rf_hp.get("chemically_informed_ratio", None) is not None:
            return None
    return rf_hp


def hp_summary_str(trial_id: int, current_best: BestTrial, parameters: dict) -> str:
    hp_summary = f"Trial {trial_id + 1:>3} |"
    for k, v in parameters.items():
        fmt_val = format(v, ".3f" if isinstance(v, float) else "")
        hp_summary += f" {k:^7}: {fmt_val:^7} |"
    try:
        energy_error = current_best.log.get_metric("energy_MAE", DataSplit.VALIDATION)
        forces_error = current_best.log.get_metric("forces_MAE", DataSplit.VALIDATION)
    except KeyError:
        energy_error = current_best.log.get_metric("energy_MAE", DataSplit.TRAIN)
        forces_error = current_best.log.get_metric("forces_MAE", DataSplit.TRAIN)

    hp_summary += (
        f" Best trial {current_best.trial_id} (energy {energy_error:.2f} meV/atom - "
        f"forces {forces_error:.1f} meV/Ang)"
    )
    return hp_summary


def run_autotune(
    rnd_seed: int,
    gnn_backbone_id: str,
    interaction_block: int,
    rf_type: str,
    loaders: Dict[str, torch.utils.data.DataLoader],
    rf_param_grid,
    solver_param_grid,
    scale_by_Z: bool,
    atomic_energies: Dict[int, torch.Tensor],
    jac_chunk_size: int | Literal["auto"],
    trainer: BaseTrainer,
):
    current_best = BestTrial(None, None)
    for trial_id, parameters_instance in rf_param_grid:
        logger.debug(f"Autotune iteration with RF parameters {parameters_instance}")
        if rf_type in ["poly", "gaussian", "biased-gaussian", "multiscale-gaussian"]:
            random_features_params = {"rng_seed": rnd_seed} | parameters_instance
        elif rf_type == "linear":
            random_features_params = parameters_instance
        else:
            raise ValueError(rf_type)

        model = FrankenPotential(
            gnn_backbone_id,
            rf_type,
            random_features_params,
            interaction_block=interaction_block,
            scale_by_Z=scale_by_Z,  # TODO: This should be a param we can do hp-search on
            num_species=loaders["train"].dataset.num_species,
            atomic_energies=atomic_energies,
            jac_chunk_size=jac_chunk_size,
        )

        logs, weights = trainer.fit(model, solver_param_grid)
        for split_name, loader in loaders.items():
            logs = trainer.evaluate(model, loader, logs, weights)
        split_for_best_model = (
            DataSplit.VALIDATION if "val" in loaders else DataSplit.TRAIN
        )
        if dist_utils.get_rank() == 0:
            if trainer.log_dir is not None:
                trainer.serialize_logs(model, logs, weights, split_for_best_model)
        dist_utils.barrier()

        # current best model update
        if dist_utils.get_rank() == 0:
            if trainer.log_dir is not None:
                with open(trainer.log_dir / "best.json", "r") as f:
                    try:
                        best_log = LogEntry.from_dict(json.load(f))
                        if best_log != current_best.log:
                            current_best = BestTrial(
                                trial_id=trial_id + 1,
                                log=best_log,
                            )
                    except KeyError:
                        pass

                logger.info(hp_summary_str(trial_id, current_best, parameters_instance))
        garbage_collection_cuda()


def create_run_folder(base_run_dir: Path) -> Path:
    # Ensure the base directory exists
    if not base_run_dir.exists():
        raise FileNotFoundError(
            f"The base run directory '{base_run_dir}' does not exist."
        )
    # Use time + a shortened UUID to ensure uniqueness of
    # the experiment directory. The names will look like
    # 'run_240926_113513_1d93b3ed'
    now_str = datetime.datetime.now().strftime("%y%m%d_%H%M%S")
    exp_dir_name = f"run_{now_str}_{uuid4().hex[:8]}"
    exp_dir = base_run_dir / exp_dir_name
    exp_dir.mkdir(parents=True, exist_ok=True)
    return exp_dir


def save_experiment_info(cfg: DictConfig, run_dir: Path):
    train_hardware = {
        "num_gpus": dist_utils.get_world_size(),
        "gpu_model": get_device_name("cuda:0"),
        "cpu_model": get_device_name("cpu"),
    }
    cfg_dict = OmegaConf.to_container(cfg, resolve=True)
    assert isinstance(cfg_dict, dict)
    with open(run_dir / "configs.json", "w") as f:
        json.dump(cfg_dict | train_hardware, f, indent=4)


def get_dataset_paths(
    train_path: Optional[str],
    val_path: Optional[str],
    test_path: Optional[str],
    dataset_name: Optional[str],
) -> Tuple[Path, Optional[Path], Optional[Path]]:
    if dataset_name is not None and dataset_name in DATASET_REGISTRY.keys():
        # Load from the dataset registry
        out_train_path = DATASETS_PATH / DATASET_REGISTRY[dataset_name]["train"]
        out_val_path = DATASET_REGISTRY[dataset_name].get("val", None)
        if out_val_path is not None:
            out_val_path = DATASETS_PATH / out_val_path
        out_test_path = DATASET_REGISTRY[dataset_name].get("test", None)
        if out_test_path is not None:
            out_test_path = DATASETS_PATH / out_test_path
    else:
        if train_path is None:
            raise ValueError(
                "Either 'dataset_name' or 'train_path' must be "
                "specified in order to load a training dataset."
            )
        out_train_path = Path(train_path)
        out_val_path = Path(val_path) if val_path is not None else None
        out_test_path = Path(test_path) if test_path is not None else None
    return out_train_path, out_val_path, out_test_path


def main(cfg: DictConfig):
    run_dir = Path(cfg.paths.run_dir)

    if torch.cuda.is_available():
        rank = dist_utils.init(distributed=cfg.distributed)
        device = torch.device(torch.cuda.current_device())
    else:
        rank = 0
        device = torch.device("cpu")

    if rank != 0:  # first rank goes forward
        run_dir = None
        dist_utils.barrier()
    else:
        run_dir = create_run_folder(run_dir)
        dist_utils.barrier()  # other ranks follow
    logging_level = cfg.get("console_logging_level", "WARNING").upper()
    setup_logger(
        level=logging_level, directory=dist_utils.broadcast_obj(run_dir), rank=rank
    )
    pprint_config(cfg)

    # Global try-catch after setup_logger, to log any exceptions.
    try:
        CacheDir.initialize()

        if rank != 0:  # first rank goes forward
            dist_utils.barrier()
        else:
            save_experiment_info(cfg, run_dir)
            # Download the checkpoint if not present locally
            download_checkpoint(cfg.franken.gnn_backbone_id)
            logger.info(f"Run folder: {run_dir}")
            dist_utils.barrier()

        train_path, val_path, test_path = get_dataset_paths(
            cfg.dataset.get("train_path", None),
            cfg.dataset.get("val_path", None),
            cfg.dataset.get("test_path", None),
            cfg.dataset.get("dataset_name", None),
        )

        loaders = init_loaders(
            cfg.franken.gnn_backbone_id,
            train_path,
            val_path,
            test_path,
            cfg.dataset.train_subsample.num,
            cfg.dataset.train_subsample.rng_seed,
        )

        # Create HP sweep with "random_features" and "solver" groups.
        sweep = {}
        for group_name, group in cfg.hyperparameters.items():
            sweep[group_name] = create_params_grid(group_name, group)

        trainer: BaseTrainer = hydra.utils.instantiate(
            cfg.trainer,
            train_dataloader=loaders["train"],
            log_dir=run_dir,
            device=device,
        )

        if device.type == "cuda":
            from franken.backbones.wrappers.common_patches import patch_e3nn

            patch_e3nn()

        run_autotune(
            rnd_seed=cfg.seed,
            gnn_backbone_id=cfg.franken.gnn_backbone_id,
            interaction_block=cfg.franken.interaction_block,
            rf_type=cfg.franken.kernel_type,
            loaders=loaders,
            rf_param_grid=sweep["random_features"],
            solver_param_grid=sweep["solver"],
            scale_by_Z=cfg.franken.scale_by_Z,
            atomic_energies=cfg.franken.atomic_energies,
            jac_chunk_size=cfg.franken.jac_chunk_size,
            trainer=trainer,
        )
    except Exception as e:
        logger.error("Error encountered in autotune. Exiting.", exc_info=e)
        raise
    finally:
        if torch.distributed.is_initialized():
            torch.distributed.destroy_process_group()

    return run_dir


@hydra.main(version_base="1.3", config_path="configs", config_name="master.yaml")
def autotune(cfg: DictConfig):
    _ = main(cfg)


def cli_entry_point():
    autotune()


if __name__ == "__main__":
    autotune()



import re
import argparse
from ase.io import read


def parse_configuration(data):
    # Parse sections using regular expressions
    num_atoms = int(re.search(r"The number of atoms\s*[-=]+\s*(\d+)", data).group(1))
    energy = float(
        re.search(r"Total energy \(eV\)\s*[-=]+\s*([-+]?\d*\.\d+|\d+)", data).group(1)
    )

    # Extract lattice vectors
    lattice_match = re.search(
        r"Primitive lattice vectors \(ang.\)\s*[-=]+\s*([\d\s.-]+)", data
    )
    lattice_lines = lattice_match.group(1).strip().split("\n")
    lattice = [line.split() for line in lattice_lines]

    # Flatten and format lattice as a string for XYZ format
    lattice_flat = " ".join([" ".join(line) for line in lattice])

    # Extract atomic positions
    positions_match = re.search(
        r"Atomic positions \(ang.\)\s*[-=]+\s*([\d\s.-]+)", data
    )
    positions_lines = positions_match.group(1).strip().split("\n")
    positions = [line.split() for line in positions_lines]

    # Extract forces
    forces_match = re.search(r"Forces \(eV ang.\^-1\)\s*[-=]+\s*([\d\s.-]+)", data)
    forces_lines = forces_match.group(1).strip().split("\n")
    forces = [line.split() for line in forces_lines]

    # Extract stress tensor (two lines) without separators
    stress_match_1 = re.search(
        r"Stress \(kbar\)\s*[-=]+\s*XX YY ZZ\s*[-=]+\s*([\d\s.-]+)", data
    )
    stress_match_2 = re.search(r"XY YZ ZX\s*[-=]+\s*([\d\s.-]+)", data)

    # Ensure we only capture numerical values and not separator lines
    stress_values_1 = (
        stress_match_1.group(1).strip().split()[:3]
    )  # Take first three values for XX YY ZZ
    stress_values_2 = (
        stress_match_2.group(1).strip().split()[:3]
    )  # Take first three values for XY YZ ZX
    xx, yy, zz = stress_values_1
    xy, yz, zx = stress_values_2

    # Combine the two stress components into a single list
    # stress_tensor = stress_values_1 + stress_values_2
    # stress_tensor = ' '.join(stress_tensor)  # Convert to a single string
    stress_tensor = f"{xx} {xy} {zx} {xy} {yy} {yz} {zx} {yz} {zz}"

    # Create the extended XYZ content for this configuration
    xyz_content = []
    xyz_content.append(f"{num_atoms}")
    xyz_content.append(
        f'Lattice="{lattice_flat}" Properties=species:S:1:pos:R:3:forces:R:3 energy={energy} stress="{stress_tensor}"'
    )

    # Atom types (order them according to the positions provided)
    atom_type_lines = (
        re.search(r"Atom types and atom numbers\s*[-=]+\s*([\w\s\d]+)", data)
        .group(1)
        .strip()
        .split("\n")
    )
    atom_types = []
    for line in atom_type_lines:
        element, count = line.split()
        atom_types.extend([element] * int(count))

    # Add each atom's data line by line
    for idx, (position, force) in enumerate(zip(positions, forces)):
        element = atom_types[idx]
        px, py, pz = position
        fx, fy, fz = force
        xyz_content.append(f"{element} {px} {py} {pz} {fx} {fy} {fz}")

    return "\n".join(xyz_content)


def parse_multiple_configurations(data):
    # Split the data by configurations using "Configuration num." as the delimiter
    configurations = re.split(r"Configuration num\.\s*\d+", data)
    xyz_all = []

    # Process each configuration if it is not empty
    for config in configurations:
        config = config.strip()
        if config:  # Only parse if the configuration is not empty
            try:
                xyz_all.append(parse_configuration(config))
            except Exception as e:
                print(f"Error parsing configuration: {e}")

    # Join all configurations with a newline
    return "\n".join(xyz_all)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="MLFF to xyz converter")
    parser.add_argument("-f", "--filename", help="Input Filename")
    args = parser.parse_args()

    filename = args.filename

    print(f"Conversion input: {filename}")

    with open(filename, "r") as file:
        data = file.read()

    xyz_output = parse_multiple_configurations(data)

    with open(filename + ".xyz", "w") as file:
        file.write(xyz_output)

    print(f"Output written to {filename}.xyz")

    traj = read(filename + ".xyz", index=":")
    print(f"Dataset has {len(traj)} frames")
    for i, atoms in enumerate(traj):
        atoms.get_potential_energy()
        atoms.get_forces()

    print("End of conversion")



import os
import argparse
from typing import Optional
import torch
from e3nn.util import jit

from franken.data.base import Configuration
from franken.rf.model import FrankenPotential


@jit.compile_mode("script")
class LammpsFrankenCalculator(torch.nn.Module):
    def __init__(
        self,
        franken_model: FrankenPotential,
    ):
        super().__init__()

        self.model = franken_model
        self.register_buffer("atomic_numbers", self.model.gnn.base_model.atomic_numbers)
        self.register_buffer("r_max", self.model.gnn.base_model.r_max)
        self.register_buffer(
            "num_interactions", self.model.gnn.base_model.num_interactions
        )
        # this attribute is used for dtype detection in LAMMPS-MACE.
        # See: https://github.com/ACEsuit/lammps/blob/mace/src/ML-MACE/pair_mace.cpp#314
        self.model.node_embedding = self.model.gnn.base_model.node_embedding

        for param in self.model.parameters():
            param.requires_grad = False

    def forward(
        self,
        data: dict[str, torch.Tensor],
        local_or_ghost: torch.Tensor,
        compute_virials: bool = False,
    ) -> dict[str, torch.Tensor | None]:
        # node_attrs is a one-hot representation of the atom types
        atom_nums = torch.nonzero(data["node_attrs"])[:, 1] + 1
        franken_data = Configuration(
            atom_pos=data["positions"],
            atomic_numbers=atom_nums,
            natoms=torch.tensor(
                len(atom_nums), dtype=torch.int32, device=atom_nums.device
            ).view(1),
            node_attrs=data["node_attrs"],
            edge_index=data["edge_index"],
            shifts=data["shifts"],
            unit_shifts=data["unit_shifts"],
        )
        energy, forces = self.model(franken_data)  # type: ignore
        # Kokkos doesn't like total_energy_local and only looks at node_energy.
        # We hack around this:
        node_energy = energy.repeat(len(atom_nums)).div(len(atom_nums))
        virials: Optional[torch.Tensor] = None
        if compute_virials:
            virials = torch.zeros((1, 3, 3), dtype=forces.dtype, device=forces.device)
        return {
            "total_energy_local": energy,
            "node_energy": node_energy,
            "forces": forces,
            "virials": virials,
        }


def create_lammps_model():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--model_path",
        type=str,
        help="Path to the model to be converted to LAMMPS",
    )
    parser.add_argument(
        "--rf_weight_id",
        type=int,
        help="Head of the model to be converted to LAMMPS",
        default=None,
    )
    args = parser.parse_args()

    franken_model = FrankenPotential.load(
        args.model_path,
        map_location=torch.device("cuda" if torch.cuda.is_available() else "cpu"),
        rf_weight_id=args.rf_weight_id,
    )
    # NOTE:
    # Kokkos is hardcoded to double and will silently corrupt data if the model
    # does not use dtype double.
    franken_model = franken_model.double().to("cpu")
    lammps_model = LammpsFrankenCalculator(franken_model)
    lammps_model_compiled = jit.compile(lammps_model)

    save_path = f"{os.path.splitext(args.model_path)[0]}-lammps.pt"
    print(f"Saving compiled model to '{save_path}'")
    lammps_model_compiled.save(save_path)


if __name__ == "__main__":
    create_lammps_model()



def test_registry_init():
    from franken.metrics.registry import registry

    assert hasattr(registry._instance, "_metrics")


def test_available_metrics():
    import franken.metrics as fm

    for name in ["energy_MAE", "forces_MAE", "forces_cosim"]:
        assert name in fm.available_metrics()


def test_register():
    import franken.metrics as fm
    from franken.metrics.base import BaseMetric

    class MockMetric(BaseMetric):
        pass

    assert "mock_metric" not in fm.available_metrics()
    fm.register("mock_metric", MockMetric)
    assert "mock_metric" in fm.available_metrics()


def test_init_metric():
    import torch

    import franken.metrics as fm
    from franken.metrics.base import BaseMetric

    class MockMetric(BaseMetric):
        def __init__(self, device: torch.device, dtype: torch.dtype = torch.float32):
            super().__init__("mock_metric", device, dtype)

    fm.register("mock_metric", MockMetric)
    metric = fm.init_metric("mock_metric", torch.device("cpu"))
    assert isinstance(metric, MockMetric)
